"""Strptime-related classes and functions.

CLASSES:
    TimeRE -- Creates regexes for pattern matching a string of text containing
                time information

FUNCTIONS:
    strptime -- Calculates the time struct represented by the passed-in string

"""
import time
import calendar
from re import compile as re_compile
from re import IGNORECASE
from re import escape as re_escape
from datetime import date as datetime_date
try:
    from thread import allocate_lock as _thread_allocate_lock
except:
    from dummy_thread import allocate_lock as _thread_allocate_lock

__author__ = "Brett Cannon"
__email__ = "brett@python.org"

__all__ = ['strptime']

class TimeRE(dict):
    """Handle conversion from format directives to regexes."""

    def __init__(self):
        """Create keys/values.

        Order of execution is important for dependency reasons.

        """
        base = super(TimeRE, self)
        base.__init__({
            # The " \d" part of the regex is to make %c from ANSI C work
            'd': r"(?P<d>3[0-1]|[1-2]\d|0[1-9]|[1-9]| [1-9])",
            'H': r"(?P<H>2[0-3]|[0-1]\d|\d)",
            'I': r"(?P<I>1[0-2]|0[1-9]|[1-9])",
            'j': r"(?P<j>36[0-6]|3[0-5]\d|[1-2]\d\d|0[1-9]\d|00[1-9]|[1-9]\d|0[1-9]|[1-9])",
            'm': r"(?P<m>1[0-2]|0[1-9]|[1-9])",
            'M': r"(?P<M>[0-5]\d|\d)",
            'S': r"(?P<S>6[0-1]|[0-5]\d|\d)",
            'U': r"(?P<U>5[0-3]|[0-4]\d|\d)",
            'w': r"(?P<w>[0-6])",
            # W is set below by using 'U'
            'y': r"(?P<y>\d\d)",
            #XXX: Does 'Y' need to worry about having less or more than
            #     4 digits?
            'Y': r"(?P<Y>\d\d\d\d)",
            'A': self.__seqToRE(['monday', 'tuesday', 'wednesday', 'thursday', 'friday', 'saturday', 'sunday'], 'A'),
            'a': self.__seqToRE(['mon', 'tue', 'wed', 'thu', 'fri', 'sat', 'sun'], 'a'),
            'B': self.__seqToRE(['january', 'february', 'march', 'april', 'may', 'june', 'july', 'august', 'september', 'october', 'november', 'december'], 'B'),
            'b': self.__seqToRE(['jan', 'feb', 'mar', 'apr', 'may', 'jun', 'jul', 'aug', 'sep', 'oct', 'nov', 'dec'], 'b'),
            'p': self.__seqToRE(['am', 'pm'], 'p'),
            'Z': self.__seqToRE(['utc', 'cst', 'gmt', 'cst'], 'Z'),
            '%': '%'})
        base.__setitem__('W', base.__getitem__('U').replace('U', 'W'))
        base.__setitem__('c', self.pattern('%a %b %d %H:%M:%S %Y'))
        base.__setitem__('x', self.pattern('%m/%d/%y'))
        base.__setitem__('X', self.pattern('%H:%M:%S'))

    def __seqToRE(self, to_convert, directive):
        """Convert a list to a regex string for matching a directive.

        Want possible matching values to be from longest to shortest.  This
        prevents the possibility of a match occuring for a value that also
        a substring of a larger value that should have matched (e.g., 'abc'
        matching when 'abcdef' should have been the match).

        """
        to_convert = sorted(to_convert, key=len, reverse=True)
        for value in to_convert:
            if value != '':
                break
        else:
            return ''
        regex = '|'.join(re_escape(stuff) for stuff in to_convert)
        regex = '(?P<%s>%s' % (directive, regex)
        return '%s)' % regex

    def pattern(self, format):
        """Return regex pattern for the format string.

        Need to make sure that any characters that might be interpreted as
        regex syntax are escaped.

        """
        processed_format = ''
        # The sub() call escapes all characters that might be misconstrued
        # as regex syntax.  Cannot use re.escape since we have to deal with
        # format directives (%m, etc.).
        regex_chars = re_compile(r"([\\.^$*+?\(\){}\[\]|])")
        format = regex_chars.sub(r"\\\1", format)
        whitespace_replacement = re_compile('\s+')
        format = whitespace_replacement.sub('\s*', format)
        while '%' in format:
            directive_index = format.index('%')+1
            processed_format = "%s%s%s" % (processed_format,
                                           format[:directive_index-1],
                                           self[format[directive_index]])
            format = format[directive_index+1:]
        return "%s%s" % (processed_format, format)

    def compile(self, format):
        """Return a compiled re object for the format string."""
        return re_compile(self.pattern(format), IGNORECASE)

_cache_lock = _thread_allocate_lock()
# DO NOT modify _TimeRE_cache or _regex_cache without acquiring the cache lock
# first!
_TimeRE_cache = TimeRE()
_CACHE_MAX_SIZE = 5 # Max number of regexes stored in _regex_cache
_regex_cache = {}

def strptime(data_string, format="%a %b %d %H:%M:%S %Y"):
    """Return a time struct based on the input string and the format string."""
    global _TimeRE_cache
    _cache_lock.acquire()
    try:
        time_re = _TimeRE_cache
        if len(_regex_cache) > _CACHE_MAX_SIZE:
            _regex_cache.clear()
        format_regex = _regex_cache.get(format)
        if not format_regex:
            format_regex = time_re.compile(format)
            _regex_cache[format] = format_regex
    finally:
        _cache_lock.release()
    found = format_regex.match(data_string)
    if not found:
        raise ValueError("time data did not match format:  data=%s  fmt=%s" %
                         (data_string, format))
    if len(data_string) != found.end():
        raise ValueError("unconverted data remains: %s" %
                          data_string[found.end():])
    year = 1900
    month = day = 1
    hour = minute = second = 0
    tz = -1
    # Default to -1 to signify that values not known; not critical to have,
    # though
    week_of_year = -1
    week_of_year_start = -1
    # weekday and julian defaulted to -1 so as to signal need to calculate
    # values
    weekday = julian = -1
    found_dict = found.groupdict()
    for group_key in found_dict.iterkeys():
        # Directives not explicitly handled below:
        #   c, x, X
        #      handled by making out of other directives
        #   U, W
        #      worthless without day of the week
        if group_key == 'y':
            year = int(found_dict['y'])
            # Open Group specification for strptime() states that a %y
            #value in the range of [00, 68] is in the century 2000, while
            #[69,99] is in the century 1900
            if year <= 68:
                year += 2000
            else:
                year += 1900
        elif group_key == 'Y':
            year = int(found_dict['Y'])
        elif group_key == 'm':
            month = int(found_dict['m'])
        elif group_key == 'B':
            month = ['', 'january', 'february', 'march', 'april', 'may', 'june', 'july', 'august', 'september', 'october', 'november', 'december'].index(found_dict['B'].lower())
        elif group_key == 'b':
            month = ['', 'jan', 'feb', 'mar', 'apr', 'may', 'jun', 'jul', 'aug', 'sep', 'oct', 'nov', 'dec'].index(found_dict['b'].lower())
        elif group_key == 'd':
            day = int(found_dict['d'])
        elif group_key == 'H':
            hour = int(found_dict['H'])
        elif group_key == 'I':
            hour = int(found_dict['I'])
            ampm = found_dict.get('p', '').lower()
            # If there was no AM/PM indicator, we'll treat this like AM
            if ampm in ('', 'am'):
                # We're in AM so the hour is correct unless we're
                # looking at 12 midnight.
                # 12 midnight == 12 AM == hour 0
                if hour == 12:
                    hour = 0
            elif ampm == 'pm':
                # We're in PM so we need to add 12 to the hour unless
                # we're looking at 12 noon.
                # 12 noon == 12 PM == hour 12
                if hour != 12:
                    hour += 12
        elif group_key == 'M':
            minute = int(found_dict['M'])
        elif group_key == 'S':
            second = int(found_dict['S'])
        elif group_key == 'A':
            weekday = ['monday', 'tuesday', 'wednesday', 'thursday', 'friday', 'saturday', 'sunday'].index(found_dict['A'].lower())
        elif group_key == 'a':
            weekday = ['mon', 'tue', 'wed', 'thu', 'fri', 'sat', 'sun'].index(found_dict['a'].lower())
        elif group_key == 'w':
            weekday = int(found_dict['w'])
            if weekday == 0:
                weekday = 6
            else:
                weekday -= 1
        elif group_key == 'j':
            julian = int(found_dict['j'])
        elif group_key in ('U', 'W'):
            week_of_year = int(found_dict[group_key])
            if group_key == 'U':
                # U starts week on Sunday
                week_of_year_start = 6
            else:
                # W starts week on Monday
                week_of_year_start = 0
        elif group_key == 'Z':
            # Since -1 is default value only need to worry about setting tz if
            # it can be something other than -1.
            found_zone = found_dict['Z'].lower()
            for value, tz_values in enumerate((frozenset(['utc', 'cst', 'gmt']), frozenset(['cst']))):
                if found_zone in tz_values:
                    # Deal with bad locale setup where timezone names are the
                    # same and yet time.daylight is true; too ambiguous to
                    # be able to tell what timezone has daylight savings
                    if (time.tzname[0] == time.tzname[1] and
                       time.daylight and found_zone not in ("utc", "gmt")):
                        break
                    else:
                        tz = value
                        break
    # If we know the week of the year and what day of that week, we can figure
    # out the Julian day of the year
    # Calculations below assume 0 is a Monday
    if julian == -1 and week_of_year != -1 and weekday != -1:
        # Calculate how many days in week 0
        first_weekday = datetime_date(year, 1, 1).weekday()
        preceeding_days = 7 - first_weekday
        if preceeding_days == 7:
            preceeding_days = 0
        # Adjust for U directive so that calculations are not dependent on
        # directive used to figure out week of year
        if weekday == 6 and week_of_year_start == 6:
            week_of_year -= 1
        # If a year starts and ends on a Monday but a week is specified to
        # start on a Sunday we need to up the week to counter-balance the fact
        # that with %W that first Monday starts week 1 while with %U that is
        # week 0 and thus shifts everything by a week
        if weekday == 0 and first_weekday == 0 and week_of_year_start == 6:
            week_of_year += 1
        # If in week 0, then just figure out how many days from Jan 1 to day of
        # week specified, else calculate by multiplying week of year by 7,
        # adding in days in week 0, and the number of days from Monday to the
        # day of the week
        if week_of_year == 0:
            julian = 1 + weekday - first_weekday
        else:
            days_to_week = preceeding_days + (7 * (week_of_year - 1))
            julian = 1 + days_to_week + weekday
    # Cannot pre-calculate datetime_date() since can change in Julian
    #calculation and thus could have different value for the day of the week
    #calculation
    if julian == -1:
        # Need to add 1 to result since first day of the year is 1, not 0.
        julian = datetime_date(year, month, day).toordinal() - \
                  datetime_date(year, 1, 1).toordinal() + 1
    else:  # Assume that if they bothered to include Julian day it will
           #be accurate
        datetime_result = datetime_date.fromordinal((julian - 1) + datetime_date(year, 1, 1).toordinal())
        year = datetime_result.year
        month = datetime_result.month
        day = datetime_result.day
    if weekday == -1:
        weekday = datetime_date(year, month, day).weekday()
    return time.struct_time((year, month, day, hour, minute, second, weekday, julian, tz))



import threading
from decorator import decorator

@decorator
def thread_check_callable(func, self, *args, **kw):
	assert self._thread is threading.currentThread(), "%s can only be called by %s not %s" % (func, self._thread, threading.currentThread())
	return func(self, *args, **kw)

@decorator
def thread_check_init(func, self, *args, **kw):
	self._thread = threading.currentThread()
	return func(self, *args, **kw)

def thread_safe(func):
	"""
	Mark this function as thread safe so can be called accross thread
	boundaries.  
	"""
	func.__threadsafe = True
	return func

# FIXME: This doesn't also check the classes base class!
import inspect
class thread_checker(type):
	"""
	This class can be set as the metaclass for an object to make sure that it
	is only access from a single thread.  

	Methods can be marked "thread_safe" with the thread_safe decorator also 
	found in this module.
	"""
	def __new__(cls,classname,bases,classdict):
		for attr, item in classdict.items():
			if attr == "__init__":
				classdict[attr] = thread_check_init(item)
			elif inspect.isfunction(item) and not hasattr(item, "__threadsafe"):
				classdict[attr] = thread_check_callable(item)

		t = type.__new__(cls,classname,bases,classdict)
		return t

__all__ = ["thread_checker", "thread_safe"]

if __name__ == "__main__":
	class B(object):
		__metaclass__ = thread_checker

		def __init__(self):
			pass

		test_simple  = 2
		test_complex = [1, 3]

		def test_unsafe(self):
			return "test_unsafe worked"

		@thread_safe
		def test_safe(self):
			return "test_safe worked"

	class t1(threading.Thread):
		def __init__(self, ishouldnotaccess):
			self.i = ishouldnotaccess

			threading.Thread.__init__(self)

		def run(self):
			print self.i.test_simple
			print self.i.test_complex
			print self.i.test_safe()
			print self.i.test_unsafe()

	ot = t1(B())

	print ot.i.test_safe
	print inspect.getargspec(ot.i.test_safe)
	print ot.i.test_unsafe	
	print inspect.getargspec(ot.i.test_unsafe)
	
	ot.run()

	# this should fail
	ot.start()

	import time
	time.sleep(10)




#!/usr/bin/env python
"""Harness to run a bunch of tests."""
import unittest

def suite():
    moduleNames = ['pair',
                   'parser',
                   'pogo',
                   'test_scheme',
                   'test_analyzer',
                   'expressions',
                   'expander',
                   ]

    suite = unittest.TestSuite()
    loader = unittest.defaultTestLoader
    for n in moduleNames:
        testCase = loader.loadTestsFromModule(__import__(n))
        suite.addTest(testCase)
    return suite


if __name__ == '__main__':
    suite = suite()
    runner = unittest.TextTestRunner()
    runner.run(suite)



from types import TupleType

class ChangeDict(dict):
	"""\
	A simple dictionary which also stores the "times" an object was last updated.

	Used so that we only upload/download items which have changed.
	"""
	def __init__(self):
		dict.__init__(self)
		self.times = {}
	
	def __setitem__(self, key, value):
		"""\
		This set item is special, it only takes keys of the form,
		(<last modified time>, <normal key>)
		"""
		if type(value) is TupleType and len(value) == 2:
			time, value = value
		else:
			time = -1

		if time != -1 and self.times.has_key(key) and self.times[key] > time:
			raise ValueError("The object isn't new enough to update the dictionary with! Current %s, update %s" % (self.times[key], time))
		
		self.times[key] = time
		dict.__setitem__(self, key, value)

	def __delitem__(self, key):
		del self.times[key]
		dict.__delitem__(self, key)

ChangeDict.__repr__ = dict.__repr__



"""A evaluator based on CPS form and trampolining.



FIXME: add a lot more documentation here about how this all works.


The core forms that the evaluator recognizes is the following:

    self evaluating expressions
    variable references
    QUOTE
    SET!
    DEFINE
    IF
    LAMBDA
    BEGIN
    function application

Other special forms are handled by derivation: the expander module translates
the derived forms into these core forms.

FIXME: quasiquotation hasn't been pushed off into the expander yet.

"""

__license__ = "MIT License"

import pogo
import expressions
import environment
import exceptions
import traceback
import types
import pair
from parser import parse
from symbol import false, Symbol
from error import SchemeError

def identity(val):
    """The identity function."""
    return val



"""We need to save Python's apply() function, since we use it later on
to evaluate primitive procedure calls.  For symmetry, I'm also saving
Python's eval() function within evalInUnderlyingPython()."""
evalInUnderlyingPython = eval
applyInUnderlyingPython = apply



######################################################################
## The heart of the interpreter is eval/apply.
##
## Some notes here: I'm using trampolined style to get around Python's
## lack of tail recursion.


def eval(exp, env):
    return pogo.pogo(teval(exp, env, pogo.land))


def teval(exp, env, cont):
    """Evaluates an expression 'exp' in an environment 'env'.

    Exercise 4.3 asks us to rewrite this in a more natural
    data-directed manner.  Pychecker, also, doesn't like seeing
    so many 'return' statements in one function.  *grin*
    """
    if expressions.isSelfEvaluating(exp):
        return pogo.bounce(cont, exp)
    if expressions.isVariable(exp):
        return pogo.bounce(cont, environment.lookupVariableValue(exp, env))
    if expressions.isQuoted(exp):
        return evalQuoted(exp, env, cont)
    if expressions.isAssignment(exp):
        return evalAssignment(exp, env, cont)
    if expressions.isDefinition(exp):
        return evalDefinition(exp, env, cont)
    if expressions.isIf(exp):
        return evalIf(exp, env, cont)
    if expressions.isLambda(exp):
        return pogo.bounce(cont, expressions.makeProcedure
                          (expressions.lambdaParameters(exp),
                           expressions.lambdaBody(exp),
                           env))
    if expressions.isBegin(exp):
        return evalSequence(expressions.beginActions(exp), env, cont)
    if expressions.isApplication(exp):
        return evalApplication(exp, env, cont)
    raise SchemeError, "Unknown expression type -- eval " + str(exp)



def apply(procedure, arguments, env, cont):
    """Applies a procedure on a list of arguments."""
    if expressions.isPrimitiveProcedure(procedure):
        return applyPrimitiveProcedure(procedure, arguments, env, cont)
    elif expressions.isContinuationProcedure(procedure):
        return applyContinuationProcedure(procedure, arguments)
    if expressions.isCompoundProcedure(procedure):
        newEnv = environment.extendEnvironment(
            expressions.procedureParameters(procedure),
            arguments,
            expressions.procedureEnvironment(procedure))
        return evalSequence(expressions.procedureBody(procedure), newEnv, cont)
    raise SchemeError, "Unknown procedure type -- apply " + str(procedure)



def applyPrimitiveProcedure(proc, args, env, cont):
    try:
        return applyInUnderlyingPython(expressions.primitiveImplementation(proc),
                                       [cont, env, pair.toPythonList(args)])
    except Exception, e:
        if isinstance(e, exceptions.SystemExit): raise e
        raise SchemeError, e


def applyContinuationProcedure(proc, args):
    try:
        return applyInUnderlyingPython(expressions.continuationImplementation(proc),
                                       pair.toPythonList(args))
    except Exception, e:
        if isinstance(e, exceptions.SystemExit): raise e
        raise SchemeError, e



def evalRands(exps, env, cont):
    """Given a list of expressions, returns a new list containing the
    values of evaluating each on of them.  If the continuation is
    given, then calls cont() on the evaluated operands instead."""
    def c1(head_val):
        def c2(tail_val):
            return pogo.bounce(cont, pair.cons(head_val, tail_val))
        return evalRands(expressions.restOperands(exps), env, c2)
    if expressions.isNoOperands(exps):
        return pogo.bounce(cont, pair.list())
    return teval(expressions.firstOperand(exps), env, c1)


def evalIf(exp, env, cont):
    def c(predicate_val):
        if expressions.isTrue(predicate_val):
            return teval(expressions.ifConsequent(exp), env, cont)
        else:
            return teval(expressions.ifAlternative(exp), env, cont)
    return teval(expressions.ifPredicate(exp), env, c)


def evalSequence(exps, env, cont):
    def c(val):
        return evalSequence(expressions.restExps(exps), env, cont)
    if expressions.isLastExp(exps):
        return teval(expressions.firstExp(exps), env, cont)
    else:
        return teval(expressions.firstExp(exps), env, c)


def evalAssignment(exp, env, cont):
    def c(val):
        environment.setVariableValue(expressions.assignmentVariable(exp),
                                     val, env)
        return pogo.bounce(cont, Symbol("ok"))
    return teval(expressions.assignmentValue(exp), env, c)


def evalDefinition(exp, env, cont):
    def c(val):
        environment.defineVariable(
            expressions.definitionVariable(exp), val, env)
        return pogo.bounce(cont, Symbol("ok"))
    return teval(expressions.definitionValue(exp), env, c)


def evalApplication(exp, env, cont):
    def c1(operator_val):
        def c2(operands_val):
            return apply(operator_val, operands_val, env, cont)
        return evalRands(expressions.operands(exp), env, c2)
    return teval(expressions.operator(exp), env, c1)


def evalQuoted(exp, env, cont):
    """Returns a quoted expression, using deepQuotedEval to look for
    UNQUOTE.

    Consequently, quoted elements are properly turned into cons pairs.
    """
    text = expressions.textOfQuotation(exp)
    if expressions.isQuasiquoted(exp):
        expandedExp = expressions.expandQuasiquotation(text)
        return pogo.bounce(teval,
                               expandedExp,
                               env,
                               cont)
    else:
        return pogo.bounce(cont, text)



#! /bin/python
"""

This file includes an implimentation of a datastructure that can handle all the
problems associated with the evil slot interface we use in Thousand Parsec.

"""


class ChangeNode(object):
	NodeCounter = 0

	states = ["creating", "idle", "updating", "removing", "removed"]

	def __init__(self, what):
		# FIXME: Must be a better way to do this...
		ChangeNode.NodeCounter += 1
		self.id = ChangeNode.NodeCounter

		self.left  = None
		self.right = None

		self._what    = what
		self._pending = []

	def AddState(self, state, pending=None):
		if self.LastState in ("removing", "removed"):
			raise SystemError("Can not add new states to the node if it is being removed!")

		self._pending.append([state, pending])

	def UpdatePending(self, pending):
		assert len(self._pending) > 0

		self._pending[0] = (self._pending[0][0], pending)

	@property
	def LastState(self):
		if len(self._pending) == 0:
			return "idle"
		else:
			return self._pending[-1][0]

	@property
	def CurrentState(self):
		if len(self._pending) == 0:
			return "idle"
		else:
			return self._pending[0][0]

	@property
	def ServerOrder(self):
		"""\
		The value as it currently exists on the server.
		"""
		return self._what

	@property
	def PendingOrder(self):
		"""\
		Returns the first pending change.
		"""
		if len(self._pending) == 0:
			return self._what
		else:
			return self._pending[0][1]
	
	@property
	def CurrentOrder(self):
		"""\
		Returns the 'latest' order
		"""
		order = None
		for state, order in reversed(self._pending):
			if not order is None:
				break

		if not order is None:
			return order

		return self._what

	def PopState(self):
		assert len(self._pending) > 0

		state, pending = self._pending.pop(0)

		if state == "updating" or state == "creating":
			assert not pending is None
			self._what = pending

	def inlist(self):
		return \
			(self.left  is None or self.left.right == self) \
				and \
			(self.right is None or self.right.left == self)

	def __repr__(self):
		if self.left is None:
			l = "None"
		else:
			l = hex(self.left.id)

		if self.right is None:
			r = "None"
		else:
			r = hex(self.right.id)

		return "<Node(%x)-%r <%s %s>>" % (self.id, self._what, l, r)
	__str__ = __repr__

	def __eq__(self, other):
		try:
			return self._what is other._what
		except AttributeError:
			return False

	def __neq__(self, other):
		return not self.__eq__(other)

	@property
	def pending(self):
		assert False

	@property
	def what(self):
		assert False

class ChangeHead(ChangeNode):
	def __init__(self):
		ChangeNode.__init__(self, None)

		self.__dict__['CurrentState'] = "head"
		self.__dict__['LastState']    = "head"

	def __repr__(self):
		return ChangeNode.__repr__(self).replace("Node", "Head")
	__str__ = __repr__

	def SetState(self, *args, **kw):
		raise SystemError("Should not be calling this on a head node!")

class ChangeList(object):
	def __init__(self):
		self.head = ChangeHead()

	def __getitem__(self, id):
		if isinstance(id, slice):
			return self.__iter__(id)

		node = self.head.right
		while node != None:
			if node.id == id:
				return node
			node = node.right
		raise KeyError("No node exists with id %s" % id)

	def __delitem__(self, id):
		node = self[id]

		node.left.right = node.right
		if not node.right is None:
			node.right.left = node.left

	def __contains__(self, tofind):
		if tofind is self.head:
			return True
		try:
			self[tofind.id]
			return True
		except KeyError:
			return False

	def __len__(self):
		l = 0
		
		node = self.head.right
		while node != None:
			l += 1
			node = node.right

		return l

	def __iter__(self, sliceme=None):
		# FIXME: Should be better way to do this
		if sliceme is None:
			sliceme = slice(0, len(self), 1)

		start = sliceme.start
		stop  = sliceme.stop
		step  = sliceme.step

		if step is None:
			step  = 1
		if start is None:
			start = 0
		if stop is None:
			stop  = len(self)

		if start < 0:
			start += len(self)
		if stop < 0:
			stop  += len(self)

		i = 0

		node = self.head.right
		while node != None:
			if i >= start and i < stop and i % step == 0:
				yield node
			node = node.right

			i += 1

		raise StopIteration

	def index(self, needle):
		if needle is self.head:
			return -1
		for i, node in enumerate(self):
			if node is needle:
				return i
		raise IndexError("%s was not found in the list!" % needle)

	def slot(self, needle):
		if needle is self.head:
			return -1
		i = 0
		for node in self:
			if node is needle:
				return i
			if node.CurrentState != "creating":
				i += 1
		raise IndexError("%s was not found in the list!" % needle)

	def append(self, toappend):
		node = self.head
		while node.right != None:
			node = node.right
		self.insert_after(node, toappend)

	def find(self, what):
		node = self.head
		while node.right != None:
			if node._what is what:
				return node
			node = node.right

	@property
	def first(self):
		if self.head.right is None:
			return self.head
		return self.head.right

	@property
	def last(self):
		node = self.head
		while node.right != None:
			node = node.right
		assert not node is None
		return node

	def insert_after(self, node, ins):
		before = node
		after  = node.right

		if not node.inlist():
			before = node.right.left

		before.right = ins
		ins.left     = before

		if not after is None:
			after.left = ins
		ins.right = after

	def insert_before(self, node, ins):
		before = node.left
		after  = node

		if not node.inlist():
			after = node.left.right
		
		after.left = ins
		ins.right  = after

		if not before is None:
			before.right = ins
		ins.left = before

if __name__ == "__main__":

	l = ChangeList()

	n1 = ChangeNode(1)
	n2 = ChangeNode(2)
	nt = ChangeNode('t')

	l.insert_after(l.head, n1)
	print '--', l.head
	for i, node in enumerate(l):
		print '->', i, node
	print

	l.insert_after(n1, n2)
	print '--', l.head
	for i, node in enumerate(l):
		print '->', i, node
	print
	
	l.insert_before(n2, nt)
	print '--', l.head
	for i, node in enumerate(l):
		print '->', i, node
	print

	del l[nt.id]
	print 't', nt
	print
	print '--', l.head
	for i, node in enumerate(l):
		print '->', i, node
	print

	l.insert_after( n1, ChangeNode('>'))
	l.insert_before(n2, ChangeNode('<'))
	print '--', l.head
	for i, node in enumerate(l):
		print '->', i, node
	print

	l.insert_after( nt, ChangeNode("a"))
	l.insert_before(nt, ChangeNode("b"))
	print 't', nt
	print
	print '--', l.head
	for i, node in enumerate(l):
		print i, node
	print

	for i in l[1:-1]:
		print i
	print

	print 0,   l.index(n1), l.first
	print -1,  l.index(n2), l.last

	print
	print '--', l.head
	for i, node in enumerate(l):
		print i, node
	print
	del l[l.last.id]
	print '--', l.head
	for i, node in enumerate(l):
		print i, node
	print

	nodes = []
	for node in l:
		nodes.append(node)

	for node in nodes:
		del l[node.id]

	print '--', l.head
	for i, node in enumerate(l):
		print i, node
	print
	



"""Syntax expansion.

This is not yet a strong foundation for a macro system, but it's a
start.  I have to finish understanding R. Kent Dybvig's "Writing
Hygenic Macros in Scheme with Syntax-Case".


At the moment, this system takes expression datums and returns
expanded expression datums.  I'll probably need to rework this, since
hygenic macros appear work on intermediate "Syntax" objects.
"""


__license__ = "MIT License"


import pair
import expressions
import pogo
import symbol
import parser



class Expander:
    def __init__(self):
        self._handlers = {}
        self._special_forms = { 'quote' : self.t_expand_quote,
                                'set!' : self.t_expand_setbang,
                                'define' : self.t_expand_define,
                                'if' : self.t_expand_if,
                                'lambda' : self.t_expand_lambda,
                                'begin' : self.t_expand_begin,
                                }
        
    def get_keyword_tag(self, expr):
        """Tries to return the keyword.  If no keyword exists, returns
        None."""
        if pair.isPair(expr) and symbol.isSymbol(pair.car(expr)):
            return str(pair.car(expr)).lower()
        return None


    def install_handler(self, keyword, handler):
        """Adds a new macro expansion handler."""
        self._handlers[keyword.lower()] = handler


    def expand(self, expr):
        """Given an expression, tries to expand it and its subexpressions
        into core forms."""
        return pogo.pogo(self.t_expand(expr, pogo.land))


    def t_expand(self, expr, cont):
        """Trampolined expander."""
        if not pair.isList(expr):
            return pogo.bounce(cont, expr)
        
        handler = self._handlers.get(self.get_keyword_tag(expr), None)
        if handler:
            ## expand, and recursively call expansion again.
            return pogo.bounce(self.t_expand, handler(expr), cont)
        elif self.is_special_form(expr):
            return pogo.bounce(self.t_expand_special_form, expr, cont)
        else:
            ## We recursively expand all the subexpressions by mapping
            ## t_expand() across the elements.
            return pogo.bounce(pair.c_listMap, self.t_expand, expr, cont)


    def is_special_form(self, expr):
        """Returns True if expr looks like a core special form."""
        return self.get_keyword_tag(expr) in self._special_forms


    def t_expand_special_form(self, expr, cont):
        """The core special forms have a different set of expansion
        rules."""
        return self._special_forms[self.get_keyword_tag(expr)](expr, cont)


    def t_expand_quote(self, expr, cont):
        if expressions.isQuasiquoted(expr):
            expandedExp = expressions.expandQuasiquotation(text)
            return pogo.bounce(self.t_expand, expandedExp, cont)
        else:
            return pogo.bounce(cont, expr)


    def t_expand_setbang(self, expr, cont):
        variable = expressions.assignmentVariable(expr)
        value = expressions.assignmentValue(expr)
        def c_expanded(expandedValue):
            return pogo.bounce(
                cont, expressions.makeAssignment(variable, expandedValue))
        return pogo.bounce(self.t_expand, value, c_expanded)


    def t_expand_define(self, expr, cont):
        variable = expressions.definitionVariable(expr)
        value = expressions.definitionValue(expr)
        def c_expanded(expandedValue):
            return pogo.bounce(
                cont, expressions.makeDefinition(variable, expandedValue))
        return pogo.bounce(self.t_expand, value, c_expanded)


    def t_expand_if(self, expr, cont):
        pred = expressions.ifPredicate(expr)
        consq = expressions.ifConsequent(expr)
        altern = expressions.ifAlternative(expr)
        def c_pred(pred_value):
            def c_conseq(consequent_value):
                def c_altern(alternative_value):
                    return pogo.bounce(cont,
                                       expressions.makeIf(pred_value,
                                                          consequent_value,
                                                          alternative_value))
                return pogo.bounce(self.t_expand, altern, c_altern)
            return pogo.bounce(self.t_expand, consq, c_conseq)
        return pogo.bounce(self.t_expand, pred, c_pred)


    def t_expand_lambda(self, expr, cont):
        parameters = expressions.lambdaParameters(expr)
        body = expressions.lambdaBody(expr)
        def c_expanded(expandedBody):
            return pogo.bounce(cont, expressions.makeLambda(parameters,
                                                            expandedBody))
        return pogo.bounce(self.t_expand, body, c_expanded)


    def t_expand_begin(self, expr, cont):
        def c_expanded(expanded_actions):
            return pogo.bounce(cont,
                               expressions.sequenceToExp(expanded_actions))
        return pair.c_listMap(self.t_expand,
                              expressions.beginActions(expr),
                              c_expanded)
                              
    
######################################################################

## For convenience, we keep a singleton instance at the module level.
_instance = Expander()
expand = _instance.expand
install_handler = _instance.install_handler


######################################################################


def LET_handler(exp):
    return expressions.letToApplication(exp)


def COND_handler(exp):
    return expressions.condToIf(exp)


def and_clauses(exp):
    return pair.cdr(exp)

def makeAnd(clauses):
    return pair.cons(symbol.Symbol("AND"), clauses)    

def AND_handler(exp):
    """We have to consider three cases:

    (AND)            ===>   #f
    (AND e1)         ===>   e1
    (AND e1 e2 ...)  ===>   (IF e1 (AND e2 ...)  #f)
    """
    clauses = and_clauses(exp)
    if pair.length(clauses) == 0:
        return symbol.false
    elif pair.length(clauses) == 1:
        return pair.car(clauses)
    else:
        return expressions.makeIf(pair.car(clauses),
                                  makeAnd(pair.cdr(clauses)),
                                  symbol.false)



def makeOr(clauses):
    return pair.cons(symbol.Symbol("OR"), clauses)



def OR_handler(exp):
    """We have to consider three cases:

    (OR)            ===>   #t

    (OR e1)         ===>   e1

    (OR e1 e2 ...)  ===>   (let ((temp-val e1))
                              (IF temp-val
                                  temp-val
                                  (OR e2 ...)))

                    ===>  ((lambda (temp-val)
                              (IF temp-val
                                  temp-val
                                  (OR e2 ...)))
                           e1)
    """
    clauses = and_clauses(exp)
    if pair.length(clauses) == 0:
        return symbol.true
    elif pair.length(clauses) == 1:
        return pair.car(clauses)
    else:
        temporarySymbol = symbol.makeUniqueTemporary()
        lambdaVal = expressions.makeLambda(
            pair.list(temporarySymbol),
            pair.list(expressions.makeIf(temporarySymbol,
                                         temporarySymbol,
                                         makeOr(pair.cdr(clauses)))))
        return expressions.makeApplication(
            lambdaVal, pair.list(pair.car(clauses)))



def install_core_handlers(expanderInstance):
    expanderInstance.install_handler('AND', AND_handler)
    expanderInstance.install_handler('OR', OR_handler)
    expanderInstance.install_handler('LET', LET_handler)
    expanderInstance.install_handler('COND', COND_handler)    
    


######################################################################    

try:
    import unittest

    class ExpanderTests(unittest.TestCase):
        def setUp(self):
            self.expander = Expander()
            install_core_handlers(self.expander)
            

        def ep(self, string):
            """Expand and parse.  Typing shortcut."""
            return self.expander.expand(parser.parse(string))


        def p(self, string):
            """Parse.  Typing shortcut."""
            return parser.parse(string)

        
        def testExpansionOnEmptyCase(self):
            self.assertEquals(self.p('()'), self.ep('()'))


        def testANDExpansion(self):
            self.assertEquals(self.p("#f"), self.ep("(and)"))
            self.assertEquals(self.p("foo"), self.ep("(and foo)"))
            self.assertEquals(self.p("(if 3 4 #f)"), self.ep("(and 3 4)"))
            self.assertEquals(self.p("(if 3 (if 4 5 #f) #f)"),
                              self.ep("(and 3 4 5)"))
            self.assertEquals(3, self.ep("(and (and (and 3)))"))

        def testOrExpansion(self):
            self.assertEquals(self.p("#t"), self.ep("(or)"))
            self.assertEquals(self.p("blah"), self.ep("(or blah)"))
            self.ep("(or foo bar)")
            self.ep("(or (and 3) (and 4) (and 5))")
            ## FIXME: I have to figure out how to test this, without
            ## knowing in advance the temporary symbols used in the
            ## expansion...


        def testQuote(self):
            self.assertEquals(self.p("(quote and)"), self.ep("'and"))


        def testSetBang(self):
            self.assertEquals(self.p("(set! x 42)"), self.ep("(set! x (AND 42))"))


        def testDefine(self):
            self.assertEquals(self.p("(define x 42)"),
                              self.ep("(define x (AND 42))"))
                              

        def testIf(self):
            self.assertEquals(self.p("(if foo bar bah!)"),
                              self.ep("(if (AND foo) (AND bar) (AND bah!))"))


        def testLambda(self):
            self.assertEquals(self.p("(lambda (x) one two)"),
                              self.ep("(lambda (x) (AND one) (AND two))"))


        def testBegin(self):
            self.assertEquals(self.p("(begin one two three)"),
                              self.ep("(begin (AND one) (begin (AND two)) (AND three))"))


        def testLet(self):
            self.assertEquals(self.p("((lambda (foo bar) (+ foo bar)) boo hoo)"),
                              self.ep("""(let ((foo boo)
                                               (bar hoo))
                                              (+ foo bar))"""))


        def testCond(self):
            self.assertEquals(self.p("""
                              (define fib
                                 (lambda (x)
                                    (if (= x 0)
                                        0
                                        (if (= x 1)
                                            1
                                            (+ (fib (- x 1)
                                                    (- x 2)))))))
                                     """),
                              self.ep("""
                              (define (fib x)
                                 (cond ((= x 0) 0)
                                       ((= x 1) 1)
                                       (else (+ (fib (- x 1)
                                                     (- x 2))))))
                              """))

        ## FIXME: add more complex cases here

    if __name__ == '__main__':
        unittest.main()

except ImportError:
    pass
        




"""Provides functions to extract portions of the expression tree and
to construct new expressions.  Heavily used in eval.py"""

__license__ = "MIT License"

import sys
from sets import Set

import pair
import types
from symbol import isSymbol, false, Symbol
import pogo
import parser


def isSelfEvaluating(exp):
    """Returns True if the expression is self-evaluating."""
    return isNumber(exp) or isString(exp) or pair.isNull(exp)


def isNumber(x):
    """Returns True if we see a number."""
    return type(x) in (types.IntType, types.FloatType, types.LongType)


def isString(x):
    """Returns True if we see a string."""
    return type(x) == types.StringType



def isVariable(exp):
    """Returns true if the expression looks like a symbol."""
    return isSymbol(exp)


def isTaggedList(exp, tag):
    """Returns true if the expression is tagged with the 'tag'."""
    if pair.isList(exp) and pair.length(exp) > 0:
        return pair.car(exp) == tag
    return 0


######################################################################

def isQuoted(exp):
    return (isTaggedList(exp, Symbol("quote")) or
            isTaggedList(exp, Symbol("quasiquote")))

def makeQuoted(exp):
    return pair.list(Symbol("quote"), exp)


def isQuasiquoted(exp):
    return isTaggedList(exp, Symbol("quasiquote"))


def isUnquoted(exp):
    return isTaggedList(exp, Symbol("unquote"))


def textOfQuotation(exp):
    return pair.cadr(exp)


def textOfUnquotation(exp):
    return pair.cadr(exp)


def expandQuasiquotation(exp, depth=1):
    """Takes a quasiquoted expression and constructs a new quoted
    expression.

    FIXME: this function is SO evil.  *grin*  Must clean this up.
    """
    if not pair.isList(exp):
        return makeQuoted(exp)
    if isUnquoted(exp):
        if depth == 1:
            return textOfUnquotation(exp)
        else:
            return pair.list(
                Symbol("list"),
                makeQuoted(Symbol("unquote")),
                expandQuasiquotation(textOfUnquotation(exp), depth - 1))
    if isQuasiquoted(exp):
        return pair.list(
            Symbol("list"),
            makeQuoted(Symbol("quasiquote")),
            expandQuasiquotation(textOfUnquotation(exp), depth + 1))
    else:
        return pair.cons(
            Symbol("list"),
            pair.listMap(lambda subexp:
                         expandQuasiquotation(subexp, depth),
                         exp))



######################################################################

def isLet(exp):
    return isTaggedList(exp, Symbol("let"))


def letBindings(exp):
    return pair.cadr(exp)


def letBindingVariables(bindings):
    if pair.isNull(bindings): return pair.list()
    return pair.cons(pair.car(pair.car(bindings)),
                           letBindingVariables(pair.cdr(bindings)))

def letBindingValues(bindings):
    if pair.isNull(bindings): return pair.list()
    return pair.cons(pair.cadr(pair.car(bindings)),
                           letBindingValues(pair.cdr(bindings)))

def letBody(exp):
    return pair.cdr(pair.cdr(exp))

def letToApplication(exp):
    bindings = letBindings(exp)
    return pair.cons(makeLambda(letBindingVariables(bindings),
                                      letBody(exp)),
                      letBindingValues(bindings))

######################################################################


def isAssignment(exp):
    """Assignments have the form (set! <var> <value>)."""
    return isTaggedList(exp, Symbol("set!"))


def assignmentVariable(exp):
    return pair.cadr(exp)


def assignmentValue(exp):
    return pair.caddr(exp)


def makeAssignment(var, val):
    return pair.list(Symbol("set!"), var, val)


######################################################################


def isDefinition(exp):
    return isTaggedList(exp, Symbol("define"))


def definitionVariable(exp):
    if isSymbol(pair.cadr(exp)):
        return pair.cadr(exp)
    return pair.car(pair.cadr(exp))


def definitionValue(exp):
    if isSymbol(pair.cadr(exp)):
        return pair.caddr(exp)
    return makeLambda(pair.cdr(pair.cadr(exp)),
                      pair.cddr(exp))

def makeDefinition(var, val):
    return pair.list(Symbol("define"), var, val)

######################################################################

def isLambda(exp):
    return isTaggedList(exp, Symbol("lambda"))


def lambdaParameters(exp):
    return pair.cadr(exp)


def lambdaBody(exp):
    return pair.cddr(exp)


def makeLambda(parameters, body):
    return pair.append(pair.list(Symbol("lambda"), parameters),
                       body)


######################################################################


def isIf(exp):
    return isTaggedList(exp, Symbol("if"))


def ifPredicate(exp):
    return pair.cadr(exp)


def ifConsequent(exp):
    return pair.caddr(exp)


def ifAlternative(exp):
    if not pair.isNull(pair.cdddr(exp)):
        return pair.cadddr(exp)
    return false


def makeIf(predicate, consequent, alternative):
    return pair.list(Symbol("if"), predicate, consequent, alternative)


######################################################################

def isBegin(exp):
    return isTaggedList(exp, Symbol("begin"))


def beginActions(exp):
    return pair.cdr(exp)


def isLastExp(seq):
    return pair.isNull(pair.cdr(seq))


def firstExp(seq):
    return pair.car(seq)


def restExps(seq):
    return pair.cdr(seq)


def sequenceToExp(seq):
    if pair.isNull(seq): return seq
    if isLastExp(seq): return firstExp(seq)
    return makeBegin(seq)


def makeBegin(seq):
    return pair.cons(Symbol("begin"), seq)


######################################################################


def isApplication(exp):
    return pair.isList(exp)


def operator(exp):
    if pair.length(exp) == 0:
        raise SchemeError, \
              "No operator given for procedure application -- OPERATOR"
    return pair.car(exp)


def operands(exp):
    return pair.cdr(exp)


def isNoOperands(ops):
    return pair.isNull(ops)


def firstOperand(ops):
    return pair.car(ops)


def restOperands(ops):
    return pair.cdr(ops)


def makeApplication(operators, operands):
    return pair.cons(operators, operands)


######################################################################


def isCond(exp):
    return isTaggedList(exp, Symbol("cond"))


def condClauses(exp):
    return pair.cdr(exp)


def isCondElseClause(clause):
    return condPredicate(clause) == Symbol("else")


def condPredicate(clause):
    return pair.car(clause)


def condActions(clause):
    return pair.cdr(clause)


def condToIf(exp):
    return expandClauses(condClauses(exp))


def expandClauses(clauses):
    if pair.isNull(clauses):
        return false
    first = pair.car(clauses)
    rest = pair.cdr(clauses)
    if isCondElseClause(first):
        if pair.isNull(rest):
            return sequenceToExp(condActions(first))
        raise SchemeError, "else clause isn't last -- condToIf" + clauses
    return makeIf(condPredicate(first),
                  sequenceToExp(condActions(first)),
                  expandClauses(rest))

def isTrue(x):
    return (x != false)


def isFalse(x):
    return (x == false)

                                
######################################################################


def makeProcedure(parameters, body, env):
    return pair.list(Symbol("procedure"), parameters, body, env)

def isCompoundProcedure(p):
    return isTaggedList(p, Symbol("procedure"))

def procedureParameters(p):
    return pair.cadr(p)

def procedureBody(p):
    return pair.caddr(p)

def procedureEnvironment(p):
    return pair.cadddr(p)



######################################################################
"""Here's where we define the primitives of Scheme."""
def isPrimitiveProcedure(proc):
    return isTaggedList(proc, Symbol("primitive"))


def primitiveImplementation(proc):
    return pair.cadr(proc)


def makePrimitiveProcedure(proc):
    return pair.list(Symbol("primitive"), proc)



## We have to tag a continuation slightly different: applying a
## continuation abandons the current computation.
def isContinuationProcedure(proc):
    return isTaggedList(proc, Symbol("continuation"))


def continuationImplementation(proc):
    return pair.cadr(proc)


def makeContinuationProcedure(proc):
    """Procedure is defined to be a function that takes a single 'val'
    argument."""
    return pair.list(Symbol("continuation"), proc)


######################################################################

def toString(expr, quoteStrings=1):
    """Given a Scheme expression, returns a string that tries to nicely
    return it as a string.
    
    If quoteStrings is true, puts quotes around string expressions.

    Notes: this is a little tricky just because we have to account for
    loopy structures.  We keep a set of expressions that we've already
    seen, to make sure we don't retrace any steps.
    """
    seenExpressionIds = Set()

    def isLoopyPair(expr):
        return (pair.isPair(expr) and
                id(pair.cdr(expr)) in seenExpressionIds)

    def markExprAsSeen(expr):
        if not pair.isNull(expr):
            seenExpressionIds.add(id(expr))

    def t_expressionToString(expr, cont):
        """Helper function for toString: written in CPS/trampolined
        form to avoid growing control context."""
        if id(expr) in seenExpressionIds:
            return pogo.bounce(cont, "[...]")

        if pair.isNull(expr):
            return pogo.bounce(cont, "()")
        if not pair.isPair(expr):
            return t_atomExpressionToString(expr, cont)
        elif isPrimitiveProcedure(expr) or isCompoundProcedure(expr):
            return t_procedureExpressionToString(expr, cont)
        else:
            return t_pairExpressionToString(expr, cont)

    def t_atomExpressionToString(expr, cont):
        """Converts an atomic expression (string, number) to a string.
        Written in CPS/trampolined form."""
        if type(expr) is str and quoteStrings:
            return pogo.bounce(cont, "\"%s\"" % expr.replace('"', '\\"'))
        else:
            return pogo.bounce(cont, str(expr))
        
    def t_procedureExpressionToString(expr, cont):
        """Converts a procedure expression to a string."""
        if isPrimitiveProcedure(expr):
            return pogo.bounce(cont,
                                   ('(%s %s <procedure-env>)') % (
                Symbol('primitive-procedure'),
                primitiveImplementation(expr)))
        elif isCompoundProcedure(expr):
            def c_parameters(parameterString):
                def c_body(bodyString):
                    return pogo.bounce(cont,
                                           '(%s %s %s <procedure-env>)' % 
                                           (Symbol('compound-procedure'),
                                            parameterString,
                                            bodyString))
                return t_expressionToString(procedureBody(expr), c_body)
            return t_expressionToString(procedureParameters(expr),
                                        c_parameters)
        
    def t_pairExpressionToString(expr, cont):
        """Converts a pair expression to a string."""
        pieces = []
        def loop(loop_exp, loop_cont):
            if pair.isNull(loop_exp):
                return pogo.bounce(loop_cont, pieces)
            elif pair.isDottedPair(loop_exp) or isLoopyPair(loop_exp):
                def c_car(carString):
                    def c_cdr(cdrString):
                        pieces.append(carString)
                        pieces.append(".")
                        pieces.append(cdrString)
                        return pogo.bounce(loop_cont, pieces)
                    return t_expressionToString(pair.cdr(loop_exp), c_cdr)
                return t_expressionToString(pair.car(loop_exp), c_car)
            else:
                def c_car(carString):
                    pieces.append(carString)
                    return pogo.bounce(loop, pair.cdr(loop_exp), loop_cont)
                return pogo.bounce(t_expressionToString,
                                       pair.car(loop_exp),
                                       c_car)

        markExprAsSeen(expr)
        return loop(expr,
                    lambda pieces: pogo.bounce(cont,
                                                   '(' +
                                                   ' '.join(pieces) +
                                                   ')'))

    return pogo.pogo(t_expressionToString(expr, pogo.land))



######################################################################

try:
    import unittest
    class ExpressionTests(unittest.TestCase):
        def setUp(self):
            self._oldrecursionlimit = sys.getrecursionlimit()
            sys.setrecursionlimit(50)

        def tearDown(self):
            sys.setrecursionlimit(self._oldrecursionlimit)

        
        def constructEvilNestedExpression(self, n):
            result = pair.list()
            for i in xrange(n-1):
                result = pair.list(result)
            return result


        def testSimpleExpressionToString(self):
            self.assertEquals("5", toString(5))
            self.assertEquals("x", toString(Symbol("x")))


        def testSimpleListExpressionToString(self):
            self.assertEquals("(x)", toString(pair.list(Symbol("x"))))
            self.assertEquals("(1 2)", toString(pair.list(1, 2)))
            self.assertEquals("(1 (2) 3)", toString(
                pair.list(1, pair.list(2), 3)))
            
        def testDottedPairs(self):
            self.assertEquals("(1 . 2)", toString(pair.cons(1, 2)))


        def testNullList(self):
            self.assertEquals("()", toString(pair.list()))
            self.assertEquals("(() ())", toString(
                pair.list(pair.list(), pair.list())))


        def testTailRecursiveExpressionToString(self):
            n = 1000
            evilExpression = self.constructEvilNestedExpression(n)
            self.assertEquals('(' * n + ')' * n,
                              toString(evilExpression))


        def testRecursiveExpressionsDontKillUs(self):
            ## Makes sure that circular structures do not kill us.
            loopyList = pair.cons(1, 1)
            pair.setCdrBang(loopyList, loopyList)
            self.assertEquals("(1 . [...])", toString(loopyList))


        def testLargerExpression(self):
            program = parser.parse("""
     (define (factorial x)
        (if (= x 0)
             1
             (* x (factorial (- x 1)))))""")
            self.assertEquals("(define (factorial x) (if (= x 0) 1 (* x (factorial (- x 1)))))",
                              toString(program))


    if __name__ == '__main__':
        unittest.main()
except ImportError:
    pass



"""'pogo' defines primitives to do trampolined-style programming."""

import traceback

__license__ = "MIT License"


def pogo(bouncer):
    """A trampoline that bounces a single bouncer.  See:

    http://www.cs.indiana.edu/hyplan/sganz/publications/icfp99/paper.pdf
    """
    try:
        while True:
            if bouncer[0] == 'land':
                return bouncer[1]
            elif bouncer[0] == 'bounce':
                bouncer = bouncer[1](*bouncer[2])
            else:
                traceback.print_exc()
                raise TypeError, "not a bouncer"
    except TypeError:
        traceback.print_exc()
        raise TypeError, "not a bouncer"


def bounce(function, *args):
    """Returns a new trampolined value that continues bouncing on the
    trampoline."""
    return ('bounce', function, args)


def land(value):
    """Returns a new trampolined value that lands off the trampoline."""
    return ('land', value)


######################################################################
try:
    import unittest
    class PogoTest(unittest.TestCase):
        def testFactorial(self):
            def iter_fact(n):
                result = 1
                while n != 0:
                    result *= n
                    n -= 1
                return result
            def tramp_fact(n, k=1):
                if n == 0:
                    return land(k)
                return bounce(tramp_fact, n-1, k*n)
            self.assertEquals(6, pogo(tramp_fact(3)))
            self.assertEquals(6, pogo(tramp_fact(3)))
            for i in [1, 10, 100, 1000]:
                self.assertEquals(iter_fact(i), pogo(tramp_fact(i)))

        def testEvenOdd(self):
            def tramp_even(n):
                if n == 0:
                    return land(True)
                return bounce(tramp_odd, n-1)
            def tramp_odd(n):
                if n == 0:
                    return land(False)
                return bounce(tramp_even, n-1)
            for i in range(20) + [100, 1000, 10000]:
                self.assertEquals(i % 2 == 0, pogo(tramp_even(i)))
                


    if __name__ == '__main__':
        unittest.main()
except ImportError:
    pass



"""An analyzer that does syntactic analysis of an expression.

Danny Yoo (dyoo@hkn.eecs.berkeley.edu)

analyze() returns a function that can be evaluated just by calling it
with the environment and continuation.  It's based (or bastardized,
depending on your perspective... *grin*) on material in Chapter 4.1.7
of Structure and Interpretation of Computer Programs:

http://mitpress.mit.edu/sicp/full-text/book/book-Z-H-26.html#%25_sec_4.1.7

This is slightly more complicated since I'm using continuation passing
style and trampolining to get around Python's call stack.

The core forms that the analyzer recognizes is the following:

    self evaluating expressions
    variable references
    QUOTE
    SET!
    DEFINE
    IF
    LAMBDA
    BEGIN
    function application

Other special forms are handled by derivation: the expander module translates
the derived forms into these core forms.

FIXME: the analyzer itself is not in continuation passing mode.  The
analyzed expressions that it produces do use CPS.

FIXME: quasiquotation hasn't been pushed off into the expander yet.

"""


__license__ = "MIT License"

import exceptions
import traceback

import expressions
import pogo
import environment
import pair
from symbol import Symbol
from error import SchemeError

applyInUnderlyingPython = apply


def analyze(exp):
    """analyze(exp) -> lambda env, cont: ...
    Given an expression, returns a new lambda function that can be applied
    on an environment and continuation."""
    if expressions.isSelfEvaluating(exp):
        return analyzeSelfEvaluating(exp)
    if expressions.isVariable(exp):
        return analyzeVariable(exp)
    if expressions.isQuoted(exp):
        return analyzeQuoted(exp)
    if expressions.isAssignment(exp):
        return analyzeAssignment(exp)
    if expressions.isDefinition(exp):
        return analyzeDefinition(exp)
    if expressions.isIf(exp):
        return analyzeIf(exp)
    if expressions.isLambda(exp):
        return analyzeLambda(exp)
    if expressions.isBegin(exp):
        return analyzeBegin(exp)
    ## Application checking must come last, after all the special forms.
    ## have been tested.
    if expressions.isApplication(exp):
        return analyzeApplication(exp)

    ## And if we get here, bad things have happened.
    raise SchemeError, ("Unknown expression type -- eval " +
                        expressions.toString(exp))


def analyzeSelfEvaluating(exp):
    return (lambda env, cont:
            pogo.bounce(cont, exp))


def analyzeVariable(exp):
    return (lambda env, cont:
            pogo.bounce(cont, environment.lookupVariableValue(exp, env)))


def analyzeQuoted(exp):
    text = expressions.textOfQuotation(exp)
    ## the common case is to handle simple quotation.
    if not expressions.isQuasiquoted(exp):  
        return (lambda env, cont:
                pogo.bounce(cont, text))
    else:
        return analyze(expressions.expandQuasiquotation(text))


def analyzeAssignment(exp):
    varName = expressions.assignmentVariable(exp)
    analyzedValueBody = analyze(expressions.assignmentValue(exp))
    def analyzed(env, cont):
        def c(varVal):
            environment.setVariableValue(varName, varVal, env)
            return pogo.bounce(cont, Symbol("ok"))
        return pogo.bounce(analyzedValueBody, env, c)
    return analyzed
    

def analyzeDefinition(exp):
    defName = expressions.definitionVariable(exp)
    analyzedDefValue = analyze(expressions.definitionValue(exp))
    def analyzed(env, cont):
        def c(defVal):
            environment.defineVariable(defName, defVal, env)
            return pogo.bounce(cont, Symbol("ok"))
        return pogo.bounce(analyzedDefValue, env, c)
    return analyzed


def analyzeIf(exp):
    analyzedPredicate = analyze(expressions.ifPredicate(exp))
    analyzedConsequent = analyze(expressions.ifConsequent(exp))
    analyzedAlternative = analyze(expressions.ifAlternative(exp))
    def analyzed(env, cont):
        def c(predicateVal):
            if expressions.isTrue(predicateVal):
                return pogo.bounce(analyzedConsequent, env, cont)
            else:
                return pogo.bounce(analyzedAlternative, env, cont)
        return pogo.bounce(analyzedPredicate, env, c)
    return analyzed



def analyzeSequence(exps):
    def sequentially(analyzedFirst, analyzedSecond):
        def c(env, cont):
            def c_first_exec(ignoredVal):
                return pogo.bounce(analyzedSecond, env, cont)
            return pogo.bounce(analyzedFirst, env, c_first_exec)
        return c
    if pair.isNull(exps):
        raise SchemeError, "Empty sequence -- ANALYZE"

    analyzedSeqs = analyze(expressions.firstExp(exps))
    exps = expressions.restExps(exps)
    while not pair.isNull(exps):
        analyzedSeqs = sequentially(analyzedSeqs,
                                    analyze(expressions.firstExp(exps)))
        exps = expressions.restExps(exps)
    return analyzedSeqs



def analyzeLambda(exp):
    params = expressions.lambdaParameters(exp)
    analyzedBody = analyzeSequence(expressions.lambdaBody(exp))
    def analyzed(env, cont):
        return pogo.bounce(cont,
                           expressions.makeProcedure(params,
                                                     analyzedBody,
                                                     env))
    return analyzed


def analyzeBegin(exp):
    analyzedActions = analyzeSequence(expressions.beginActions(exp))
    return analyzedActions



def analyzeApplication(exp):
    analyzedOperator = analyze(expressions.operator(exp))
    analyzedOperands = analyzeOperands(expressions.operands(exp))

    def analyzed(env, cont):
        def c_operator_exec(operatorVal):
            def c_operands_exec(operandVals):
                return pogo.bounce(apply, operatorVal, operandVals, env, cont)
            return pogo.bounce(analyzedOperands, env, c_operands_exec)
        return pogo.bounce(analyzedOperator, env, c_operator_exec)
    return analyzed



def analyzeOperands(operands):
    analyzedOperands = pair.listMap(analyze, operands)
    ## Simple case: if no operands, return something that just
    ## passes NIL to its continuation
    def analyzed(env, cont):
        return execRands(analyzedOperands, env, cont)
    return analyzed


def execRands(rands, env, cont):
    """Executes each operand and constructs a new list."""
    def eval_first_cont(headVal):
        def eval_rest_cont(tailVal):
            return pogo.bounce(cont, pair.cons(headVal, tailVal))
        return execRands(expressions.restOperands(rands), env, eval_rest_cont)
    if pair.isNull(rands):
        return pogo.bounce(cont, pair.NIL)
    return texec(expressions.firstOperand(rands),
                 env, eval_first_cont)



def eval(exp, env):
    """This version of eval calls analyze, and then texec()s it
    against the environment."""
    analyzedExp = analyze(exp)
    return pogo.pogo(texec(analyzedExp, env, pogo.land))


def texec(analyzedExp, env, cont):
    """texec applies the analyzedExpression against the environment.
    Trampolined by eval(), and will be used by primitives like EVAL
    and CALL/CC."""
    return analyzedExp(env, cont)



def apply(procedure, arguments, env, cont):
    """Applies a procedure on a list of arguments."""
    if expressions.isPrimitiveProcedure(procedure):
        return applyPrimitiveProcedure(procedure, arguments, env, cont)
    elif expressions.isContinuationProcedure(procedure):
        return applyContinuationProcedure(procedure, arguments)
    elif expressions.isCompoundProcedure(procedure):
        newEnv = environment.extendEnvironment (
            expressions.procedureParameters(procedure),
            arguments,
            expressions.procedureEnvironment(procedure))
        return texec(expressions.procedureBody(procedure), newEnv, cont)
    raise SchemeError, "Unknown procedure type -- apply " + str(procedure)



def applyPrimitiveProcedure(proc, args, env, cont):
    try:
        return applyInUnderlyingPython(
            expressions.primitiveImplementation(proc),
            [cont, env, pair.toPythonList(args)])
    except Exception, e:
        if isinstance(e, exceptions.SystemExit): raise e
##        traceback.print_exc(e)
        raise SchemeError, e



def applyContinuationProcedure(proc, args):
    try:
        return applyInUnderlyingPython(
            expressions.continuationImplementation(proc),
            pair.toPythonList(args))
    except Exception, e:
        if isinstance(e, exceptions.SystemExit): raise e
        raise SchemeError, e




import pprint
import socket
import sys
import time
import traceback

from media import Media

from config import load_data, save_data
from version import version

def nop(*args, **kw):
	return

class Event(Exception):
	"""
	Base class for all events which get posted.
	"""
	def type(self):
		return self.__class__.__name__[:-5]
	type = property(type)

	def __init__(self, *args, **kw):
		self.message = ""
		Exception.__init__(self, *args, **kw)

		if self.__class__.__name__[-5:] != "Event":
			raise SystemError("All event class names must end with Event!")	

		self.time = time.time()

	def __str__(self):
		return self.__unicode__().encode('ascii', 'replace')

	def __unicode__(self):
		return unicode(self.message)

from cache import Cache
from ChangeList import ChangeNode

class Application(object):
	"""
	Container for all the applications threads and the network cache.

	Calling accross threads requires you to use the .Call method on each thread - DO NOT call directly!
	The cache can be accessed by either thread at any time - be careful.
	"""
	MediaClass  = None
	FinderClass = None
	CacheClass  = None

	def __init__(self):
		if self.CacheClass is None:
			from cache import Cache
			self.CacheClass = Cache

		try:
			import signal

			# Make sure these signals go to me, rather then a child thread..
			signal.signal(signal.SIGINT,  self.Exit)
			signal.signal(signal.SIGTERM, self.Exit)
		except ImportError:
			pass

		print self.GUIClass, self.NetworkClass, self.MediaClass, self.FinderClass
		self.gui = self.GUIClass(self)
		if not self.MediaClass is None:
			self.media = self.MediaClass(self)
		else:
			self.media = None
		if not self.FinderClass is None:
			self.finder = self.FinderClass(self)
		else:
			self.finder = None

		self.cache = None

		if hasattr(self.GUIClass, "Create"):
			self.gui.Create()
		
		# Load the Configuration
		self.ConfigLoad()

	def Run(self):
		"""\
		Set the application running.
		"""
		self.StartNetwork()

		if not self.media is None:
			self.media.start()

		if not self.finder is None:
			self.finder.start()

		self.gui.start()

	def StartNetwork(self):
		self.network = self.NetworkClass(self)
		self.network.start()

	def ConfigSave(self):
		"""\
		"""
		config = self.gui.ConfigSave()
		save_data(self.ConfigFile, config)
		
		print "Saving the config...\n" + pprint.pformat(config)

	def ConfigLoad(self):
		"""\
		"""
		config = load_data(self.ConfigFile)
		if config is None:
			config = {}
	
		self.gui.ConfigLoad(config)

	def Post(self, event, source=None):
		"""\
		Post an application wide event to every thread.
		"""
		event.source = source

		self.network.Post(event)
		self.finder.Post(event)
		self.media.Post(event)

		#print "Post", event, event.source
		#import traceback
		#traceback.print_stack()
		self.gui.Call(self.gui.Post, event)

	def Exit(self, *args, **kw):
		"""
		Exit the program.
		"""
		if hasattr(self, "closing"):
			return
		self.closing = True

		self.finder.Cleanup()
		self.network.Cleanup()
		self.media.Cleanup()
		self.gui.Cleanup()


import threading
from threadcheck import thread_checker, thread_safe

class CallThreadStop(Exception):
	pass
ThreadStop = CallThreadStop

class CallThread(threading.Thread):
	"""\
	A call thread is thread which lets you queue up functions to be called
	in the thread.

	Functions are called in the order they are queue and there is no prempting
	or other fancy stuff.
	"""
	__metaclass__ = thread_checker

	def __init__(self):
		threading.Thread.__init__(self, name=self.name)
		self.exit = False
		self.reset = False
		self.tocall = []

	@thread_safe
	def run(self):
		self._thread = threading.currentThread()

		try:
			while not self.exit:
				self.every()

				if len(self.tocall) <= 0:
					self.idle()
					continue

				method, args, kw = self.tocall.pop(0)
				try:
					method(*args, **kw)
				except CallThreadStop, e:
					self.Reset()
					self.reset = False
				except Exception, e:
					self.error(e)

		except Exception, e:
			self.error(e)

		self.Cleanup()

	def every(self):
		"""\
		Called every time th run goes around a loop.

		It is called before functions are poped of the tocall list. This mean 
		it could be used to reorganise the pending requests (or even remove
		some).

		By default it does nothing.
		"""
		pass

	def idle(self):
		"""\
		Called when there is nothing left to do. Will keep getting called until
		there is something to be done.

		The default sleeps for 100ms (should most probably sleep if you don't
		want to consume 100% of the CPU).
		"""
		time.sleep(0.1)

	def error(self, error):
		"""\
		Called when an exception occurs in a function which was called. 

		The default just prints out the traceback to stderr.
		"""
		pass

	@thread_safe
	def Reset(self):
		#del self.tocall[:]
		self.reset = True

	@thread_safe
	def Cleanup(self):
		"""\
		Ask the thread to try and exit.
		"""
		del self.tocall[:]
		self.exit = True

	@thread_safe
	def Call(self, method, *args, **kw):
		"""\
		Queue a call to method in on thread.
		"""
		self.tocall.append((method, args, kw))

	@thread_safe
	def Post(self, event):
		func = 'On' + event.type
		if hasattr(self, func):
			self.Call(getattr(self, func), event)

class NotImportantEvent(Event):
	"""\
	Not Important events are things like download progress events. They occur 
	often and if one is missed there is not huge problem.
	
	The latest NotImportantEvent is always the most up to date and if there are
	pending updates only the latest in a group should be used.
	"""
	pass

from tp.netlib import Connection, failed
from tp.netlib import objects as tpobjects
class NetworkThread(CallThread):
	"""\
	The network thread deals with talking to the server via the network.
	"""
	name = "Network"

	## These are network events
	class NetworkFailureEvent(Event):
		"""\
		Raised when the network connection fails for what ever reason.
		"""
		pass

	class NetworkFailureUserEvent(NetworkFailureEvent):
		"""\
		Raised when there was a network failure because the user does not exist.
		"""
		pass

	class NetworkFailurePasswordEvent(NetworkFailureEvent):
		"""\
		Raised when there was a network failure because the password was incorrect.
		"""
		pass

	class NetworkConnectEvent(Event):
		"""\
		Raised when the network connects to a server.
		"""
		def __init__(self, msg, features, games):
			Event.__init__(self, msg)
			self.features = features
			self.games    = games

	class NetworkAccountEvent(Event):
		"""\
		Raised when an account is successful created on a server.
		"""
		pass

	class NetworkAsyncFrameEvent(Event):
		"""\
		Raised when an async frame (such as TimeRemaining) is received.
		"""
		def __init__(self, frame):
			Event.__init__(self)

			self.frame = frame

	class NetworkTimeRemainingEvent(NetworkAsyncFrameEvent):
		"""\
		Called when an async TimeRemaining frame is received. 
		"""
		def __init__(self, frame):
			if not isinstance(frame, tpobjects.TimeRemaining):
				raise SyntaxError("NetworkTimeRemainingEvent requires a TimeRemaining frame!? (got %r)", frame)
			NetworkThread.NetworkAsyncFrameEvent.__init__(self, frame)

			self.gotat      = time.time()
			self.remaining  = frame.time	

	######################################

	def __init__(self, application):
		CallThread.__init__(self)

		self.application = application
		self.connection = Connection()

	def every(self):
		"""\
		Check's if there are any async frames pending. If so creates the correct
		events and posts them.
		"""
		try:
			self.connection.pump()

			pending = self.connection.buffered['frames-async']
			while len(pending) > 0:
				if not isinstance(pending[0], tpobjects.TimeRemaining):
					break
				frame = pending.pop(0)
				self.application.Post(self.NetworkTimeRemainingEvent(frame))
		except (AttributeError, KeyError), e:
			print e


	def error(self, error):
		traceback.print_exc()
		if isinstance(error, (IOError, socket.error)):
			s  = _(u"There was an unknown network error.\n")
			s += _("Any changes since last save have been lost.\n")
			if getattr(self.connection, 'debug', False):
				s += _("A traceback of the error was printed to the console.\n")
				print error
			print repr(s)
			self.application.Post(self.NetworkFailureEvent(s))
		else:
			raise

	def NewAccount(self, username, password, email):
		"""\
		"""
		result, message = self.connection.account(username, password, email)
		if result:
			self.application.Post(self.NetworkAccountEvent(message))
		else:
			self.application.Post(self.NetworkFailureEvent(message))

	def Connect(self, host, debug=False, callback=nop, cs="unknown"):
		"""\
		"""
		try:
			if self.connection.setup(host=host, debug=debug):
				s  = _("The client was unable to connect to the host.\n")
				s += _("This could be because the server is down or there is a problem with the network.\n")
				self.application.Post(self.NetworkFailureEvent(s))
				return False
		except (IOError, socket.error), e:
			s  = _("The client could not connect to the host.\n")
			s += _("This could be because the server is down or you mistyped the server address.\n")
			self.application.Post(self.NetworkFailureEvent(s))
			return False
		callback("connecting", "downloaded", _("Successfully connected to the host..."), amount=1)
			
		try:
			callback("connecting", "progress", _("Looking for Thousand Parsec Server..."))
			if failed(self.connection.connect(("libtpclient-py/%s.%s.%s " % version[:3])+cs)):
				raise socket.error("")
		except (IOError, socket.error), e:
			s  = _("The client connected to the host but it did not appear to be a Thousand Parsec server.\n")
			s += _("This could be because the server is down or the connection details are incorrect.\n")
			self.application.Post(self.NetworkFailureEvent(s))
			return False
		callback("connecting", "downloaded", _("Found a Thousand Parsec Server..."), amount=1)

		callback("connecting", "progress", _("Looking for supported features..."))
		features = self.connection.features()
		if failed(features):
			s  = _("The client connected to the host but it did not appear to be a Thousand Parsec server.\n")
			s += _("This could be because the server is down or the connection details are incorrect.\n")
			self.application.Post(self.NetworkFailureEvent(s))
			return False
		callback("connecting", "downloaded", _("Got the supported features..."), amount=1)

		callback("connecting", "progress", _("Looking for running games..."))
		games = self.connection.games()
		if failed(games):
			games = []
		else:
			for game in games:
				callback("connecting", "progress", _("Found %(game)s playing %(ruleset)s (%(version)s)") % {'game': game.name, 'ruleset': game.rule, 'version': game.rulever})

		callback("connecting", "downloaded", _("Got the supported features..."), amount=1)

		self.application.Post(self.NetworkConnectEvent("Connected to %s" % host, features, games))
		return 

	def ConnectTo(self, host, username, password, debug=False, callback=nop, cs="unknown"):
		"""\
		Connect to a given host using a certain username and password.
		"""
		callback("connecting", "start", _("Connecting..."))
		callback("connecting", "todownload", todownload=5)
		try:
			if self.Connect(host, debug, callback, cs) is False:
				return False
			
			callback("connecting", "progress", _("Trying to Login to the server..."))
			if failed(self.connection.login(username, password)):
				s  = _("The client connected to the host but could not login because the username of password was incorrect.\n")
				s += _("This could be because you are connecting to the wrong server or mistyped the username or password.\n")
				self.application.Post(self.NetworkFailureUserEvent(s))
				return False
			callback("connecting", "downloaded", _("Logged in okay!"), amount=1)

			# Create a new cache
			self.application.cache = self.application.CacheClass(self.application.CacheClass.key(host, username))
			return True
		finally:
			callback("connecting", "finished", "")

	def CacheUpdate(self, callback):
		try:
			callback("connecting", "alreadydone", "Already connected to the server!")
			self.application.cache.update(self.connection, callback)
			self.application.cache.save()
		except ThreadStop, e:
			pass
		except Exception, e:
			self.application.Post(self.NetworkFailureEvent(e))	
			raise

	def RequestEOT(self, callback=None):
		if callback is None:
			def callback(self, *args, **kw):
				pass

		try:
			if not hasattr(self.connection, "turnfinished"):
				print "Was unable to request turnfinished."
				return

			if failed(self.connection.turnfinished()):
				print "The request for end of turn failed."
				return
		except Exception, e:
			print e

	def OnCacheDirty(self, evt):
		"""\
		When the cache gets dirty we have to push the changes to the server.
		"""
		try:
			from cache import apply
			self.application.Post(apply(self.connection, evt, self.application.cache))
		except Exception, e:
			type, val, tb = sys.exc_info()
			sys.stderr.write("".join(traceback.format_exception(type, val, tb)))
			self.application.Post(self.NetworkFailureEvent(e))
			"There where the following errors when trying to send changes to the server:"
			"The following updates could not be made:"

class MediaThread(CallThread):
	"""\
	The media thread deals with downloading media off the internet.
	"""
	name = "Media"

	## These are network events
	class MediaFailureEvent(Event):
		"""\
		Raised when the media connection fails for what ever reason.
		"""
		pass

	class MediaDownloadEvent(Event):
		"""
		Base class for media download events.
		"""
		def __init__(self, file, progress=0, size=0, localfile=None, amount=0):
			Event.__init__(self)

			self.file      = file
			self.amount    = amount
			self.progress  = progress
			self.size      = size
			self.localfile = localfile

		def __str__(self):
			return "<%s %s>" % (self.__class__.__name__, self.file)
		__repr__ = __str__

	class MediaDownloadStartEvent(MediaDownloadEvent):
		"""\
		Posted when a piece of media is started being downloaded.
		"""
		pass

	class MediaDownloadProgressEvent(MediaDownloadEvent, NotImportantEvent):
		"""\
		Posted when a piece of media is being downloaded.
		"""
		pass

	class MediaDownloadDoneEvent(MediaDownloadEvent):
		"""\
		Posted when a piece of media has been downloaded.
		"""
		pass

	class MediaDownloadAbortEvent(MediaDownloadEvent):
		"""\
		Posted when a piece of media started downloading but was canceled.
		"""
		def __str__(self):
			return "<%s>" % (self.__class__.__name__)

	class MediaUpdateEvent(Event):
		"""\
		Posted when the media was download.
		"""
		def __init__(self, files):
			Event.__init__(self)

			self.files = files

	######################################
	def __init__(self, application):
		CallThread.__init__(self)

		self.application = application

		self.todownload = {}
		self.tostop = []
	
	def idle(self):
		if len(self.todownload) <= 0:
			CallThread.idle(self)
			return

		file, timestamp = self.todownload.iteritems().next()
		def callback(blocknum, blocksize, size, self=self, file=file, tostop=self.tostop):
			progress = min(blocknum*blocksize, size)
			if blocknum == 0:
				self.application.Post(self.MediaDownloadStartEvent(file, progress, size))
	
			self.application.Post(self.MediaDownloadProgressEvent(file, progress, size, amount=blocksize))
	
			if file in tostop:
				tostop.remove(file)
				raise self.MediaDownloadAbortEvent(file)

		try:
			localfile = self.cache.get(file, callback=callback)
			self.application.Post(self.MediaDownloadDoneEvent(file, localfile=localfile))
		except self.MediaDownloadAbortEvent, e:
			self.application.Post(e)

		del self.todownload[file]

	def error(self, error):
		if isinstance(error, (IOError, socket.error)):
			s  = _("There was an unknown network error.\n")
			s += _("Any changes since last save have been lost.\n")
			self.application.Post(self.MediaFailureEvent(s))
		raise

	@thread_safe
	def Cleanup(self):
		for file in self.todownload:
			self.tostop.append(file)
		CallThread.Cleanup(self)

	@thread_safe
	def Post(self, event):
		"""
		Post an Event the current thread.
		"""
		pass

	@thread_safe
	def StopFile(self, file):
		self.tostop.append(file)

	@thread_safe
	def GetFile(self, file):
		"""\
		Get a File, return directly or start a download.
		"""
		location = self.cache.newest(file)
		if location:
			return location
		self.todownload[file] = None

	def ConnectTo(self, host, username, debug=False):
		"""\
		ConnectTo 
		"""
		self.cache = Media("http://svn.thousandparsec.net/svn/media/client/")

		# FIXME: Hack to prevent cross thread calling - should fix the media object
		files = []
		for file in self.cache.getpossible(['png', 'gif']):
			files.append(file)
		self.application.Post(self.MediaUpdateEvent(files))

from tp.netlib.discover import LocalBrowser as LocalBrowserB
from tp.netlib.discover import RemoteBrowser as RemoteBrowserB
class LocalBrowser(LocalBrowserB, threading.Thread):
	name="LocalBrowser"

	def __init__(self, *args, **kw):
		threading.Thread.__init__(self, name=self.name)
		LocalBrowserB.__init__(self, *args, **kw)

class RemoteBrowser(RemoteBrowserB, threading.Thread):
	name="RemoteBrowser"

	def __init__(self, *args, **kw):
		threading.Thread.__init__(self, name=self.name)
		RemoteBrowserB.__init__(self, *args, **kw)
	
class FinderThread(CallThread):
	"""\
	The finder thread deals with finding games.

	It uses both Zeroconf and talks to the metaserver to get the information it
	needs.
	"""
	name="Finder"

	## These are network events
	class GameEvent(Event):
		"""
		Base class for all game found/lost events.
		"""
		def __init__(self, game):
			Event.__init__(self)

			self.game = game

	class LostGameEvent(GameEvent):
		"""\
		Raised when the finder loses a game.
		"""
		pass

	class FoundGameEvent(GameEvent):
		"""\
		Raised when the finder finds a game.
		"""
		pass
	
	class LostLocalGameEvent(FoundGameEvent):
		"""\
		Raised when the finder loses a local game.
		"""
		pass

	class FoundLocalGameEvent(FoundGameEvent):
		"""\
		Raised when the finder finds a local game.
		"""
		pass
	
	class LostRemoteGameEvent(FoundGameEvent):
		"""\
		Raised when the finder loses a remote game.
		"""
		pass

	class FoundRemoteGameEvent(FoundGameEvent):
		"""\
		Raised when the finder finds a remote game.
		"""
		pass

	class FinderErrorEvent(Event):
		"""\
		Raised when the finder has an error finding games.
		"""
		pass

	class FinderFinishedEvent(Event):
		"""\
		Raised when the finder has finished searching for new games.
		"""
		pass

	def __init__(self, application):
		CallThread.__init__(self)

		self.application = application

		self.local  = LocalBrowser()
		self.local.GameFound  = self.FoundLocalGame
		self.local.GameGone   = self.LostLocalGame

		self.remote = RemoteBrowser()
		self.remote.GameFound = self.FoundRemoteGame
		self.remote.GameGone  = self.LostRemoteGame

	@thread_safe
	def FoundLocalGame(self, game):
		self.application.Post(FinderThread.FoundLocalGameEvent(game))

	@thread_safe
	def FoundRemoteGame(self, game):
		self.application.Post(FinderThread.FoundRemoteGameEvent(game))

	@thread_safe
	def LostLocalGame(self, game):
		self.application.Post(FinderThread.LostLocalGameEvent(game))

	@thread_safe
	def LostRemoteGame(self, game):
		self.application.Post(FinderThread.LostRemoteGameEvent(game))

	@thread_safe
	def Games(self):
		"""\
		Get all the currently known games.
		"""
		return self.local.games, self.remote.games

	@thread_safe
	def Cleanup(self):
		self.local.exit()
		self.remote.exit()

	@thread_safe
	def Post(self, event):
		"""
		Post an Event the current thread.
		"""
		pass

	@thread_safe	
	def run(self):
		self._thread = threading.currentThread()

		self.local.start()
		self.remote.start()




import pyscheme as scheme

def get(list, id):
	for lid, value in list:
		if lid == id:
			return value
	return None

class DesignCalculator:
	def __init__(self, cache, design):
		self.cache = cache
		self.design = design

		self.__dirty = True

	def rank(self):
		if self.__dirty:
			ranks = {}
			for component_id, number in self.design.components:
				component = self.cache.components[component_id]
	
				for property_id, value in component.properties:
					property = self.cache.properties[property_id]
					
					if not ranks.has_key(property.rank):
						ranks[property.rank] = []
	
					if not property_id in ranks[property.rank]:
						ranks[property.rank].append(property_id)

			self.__ranks = ranks
		return self.__ranks

	def change(self, component, amount):
		"""\
		change(component, amount) -> None
		
		Changes the current design by adding/subtracting the certain amount of a component.
		"""
		self.__dirty = True

		i = 0
		while True:
			# FIXME: There should be a better way to do this.
			if i >= len(self.design.components):
				self.design.components.append([component.id, amount])
				break

			if self.design.components[i][0] == component.id:
				if isinstance(self.design.components[i], tuple):
					self.design.components[i] = list(self.design.components[i])
				self.design.components[i][1] += amount
	
				if self.design.components[i][1] < 0:
					del self.design.components[i]
				break
			i += 1

	def calculate(self):
		"""\
		calculate() -> Interpretor, Properties

		Calculates all the properties on a design. 
		Returns the Interpretor and the object with the Properties.
		"""
		i = scheme.make_interpreter()

		# Step 1 -------------------------------------
		ranks = self.rank()
		print "The order I need to calculate stuff in is,", ranks

		# Step 2 -------------------------------------
		# The object which will store the properties calculated
		class Properties(dict):
			pass

		properties = Properties()
		scheme.environment.defineVariable(scheme.symbol.Symbol('design'), properties, i.get_environment())

		# Step 3 -------------------------------------
		for rank in ranks.keys():
			for property_id in ranks[rank]:
				property = self.cache.properties[property_id]

				# Where we will store the values as calculated
				bits = []
		
				# Get all the components we contain
				for component_id, amount in self.design.components:
					# Create the component object
					component = self.cache.components[component_id] 

					# Calculate the actual value for this design
					value = get(component.properties, property_id)
					if value:
						print "Now evaluating", value
						value = i.eval(scheme.parse("""(%s design)""" % value))

						print "The value calculated for component %i was %r" % (component_id, value)
					
						for x in range(0, amount):
							bits.append(value)

				print "All the values calculated where", bits
				bits_scheme = "(list"
				for bit in bits:
					bits_scheme += " " + str(bit).replace('L', '')
				bits_scheme += ")"
				print "In scheme that is", bits_scheme
				
				total = i.eval(scheme.parse("""(let ((bits %s)) (%s design bits))""" % (bits_scheme, property.calculate)))
				value, display = scheme.pair.car(total), scheme.pair.cdr(total)

				print "In total I got '%i' which will be displayed as '%s'" % (value, display)
				properties[property.name] = (property_id, value, display)

				def t(properties, name=property.name):
					return properties[name][1]
				
				i.install_function('designtype.'+property.name, t)
				
		print "The final properties we have are", properties.items()
		return i, properties
	
	def check(self, i, properties):
		"""\
		check(Interperator, Properties) -> Valid, Feedback

		Checks the requirements of a design.

		Returns if the properties are valid and a string which has human readable feedback.
		"""
		total_okay = True
		total_feedback = []

		# Step 2, calculate the requirements for the properties
		ranks = self.rank()
		for rank in ranks.keys():
			for property_id in ranks[rank]:

				property = self.cache.properties[property_id]
				if property.requirements == '':
					print "Property with id (%i) doesn't have any requirements" % property_id
					continue
			
				print "Now checking the following requirement"
				print property.requirements
				result = i.eval(scheme.parse("""(%s design)""" % property.requirements))
				print "Result was:", result
				okay, feedback = scheme.pair.car(result), scheme.pair.cdr(result)

				if okay != scheme.symbol.Symbol('#t'):
					total_okay = False
		
				if feedback != "":
					total_feedback.append(feedback)
				
		# Step 3, calculate the requirements for the components
		for component_id, amount in self.design.components:
			component = self.cache.components[component_id]
			if component.requirements == '':
				print "Component with id (%i) doesn't have any requirements" % property_id
				continue
			
			print "Now checking the following requirement"
			print component.requirements
			result = i.eval(scheme.parse("""(%s design)""" % component.requirements))
			print "Result was:", result
			okay, feedback = scheme.pair.car(result), scheme.pair.cdr(result)

			if okay != scheme.symbol.Symbol('#t'):
				total_okay = False
		
			if feedback != "":
				total_feedback.append(feedback)

		return total_okay, "\n".join(total_feedback)

	def apply(self, properties, okay, feedback):
		"""\
		apply(Properties, 
		Apply the results returned from calculate/check to the design object.
		"""
		self.design.properties = [(x[0], x[2]) for x in properties.values()]
		self.design.feedback = feedback

		self.design.used = (-1, 0)[okay]

	def update(self):
		if self.__dirty:
			i, p = self.calculate()
			okay, reason = self.check(i, p)
			self.apply(p, okay, reason)



#!/usr/bin/env python
"""
Unit tests on PyScheme.
"""


import unittest
import symbol
from symbol import Symbol
from error import SchemeError
import pair
import sys
import parser
import scheme


class RegularInterpreterMixin:
    def setUp(self):
         ## sets the recursion limit to something fairly small to test
         ## out the the tail recursion.
        self.old_recursion_limit = sys.getrecursionlimit()
        sys.setrecursionlimit(50)
        self.interp = scheme.RegularInterpreter()

    def tearDown(self):
        sys.setrecursionlimit(self.old_recursion_limit)


    def pe(self, s):
        return self.interp.eval(parser.parse(s))
    


class BasicSchemeTests:
    """Defines a set of test cases that assumes the presence of a pe()
    method."""

    def testSimpleStuff(self):
        self.assertEquals(42, self.pe("42"))


    def testBadTypes(self):
        self.assertRaises(SchemeError, self.pe, "(* (list 1 2 3 4 5) 3)")


    def testEq(self):
        self.assertEquals(symbol.true, self.pe("""
        (begin
           (define x 3)
           (define y '(4))
           (eq? (cdr (cons x y)) y))"""))
        self.assertEquals(symbol.true, self.pe("(eq? (cdr '(1)) (cdr '(1)))"))
        self.assertEquals(symbol.false, self.pe("(eq? '(1) '(1))"))


    def testDefinition(self):
        self.assertRaises(SchemeError, self.pe, "pi") ## lookup before defn
        self.assertEquals(Symbol("ok"), self.pe("(define pi 3.1415926)"))
        self.assertEquals(3.1415926, self.pe("pi"))


    def testQuotation(self):
        self.assertEquals(pair.list(Symbol("foo"),
                                          Symbol("bar")),
                          self.pe("'(foo bar)"))

    def testQuasiquotation(self):
        self.pe("(define one 1)")
        self.pe("(define two 2)")
        self.pe("(define three 3)")
        self.assertEquals(pair.list(Symbol("one"),
                                          Symbol("two"),
                                          3),
                          self.pe("`(one two ,three)"))
        self.assertEquals(pair.list(Symbol("one"),
                                          Symbol("two"),
                                          (pair.list
                                           (Symbol("unquote"),
                                            Symbol("three")))),
                          self.pe("'(one two ,three)"))

        ## this evil case occurs in R5RS
        self.assertEquals(parser.parse("(a `(b ,(+ 1 2) ,(foo 4 d) e) f)"),
                          self.pe("`(a `(b ,(+ 1 2) ,(foo ,(+ 1 3) d) e) f)"))



    def testVarArgs(self):
        self.pe("(define (mylist . args) args)")
        self.assertEquals(parser.parse("(1 2 3)"), self.pe("(mylist 1 2 3)"))
        self.pe("""(define (list2 a b . c)
                      (list a b c))""")
        self.assertEquals(pair.list(1, 2, pair.list()),
                          self.pe("(list2 1 2)"))

    def testMinusOne(self):
        self.assertEquals(-1, self.pe("(- 1)"))

    def testApply(self):
        self.assertEquals(15, self.pe("(apply + '(1 2 3 4 5))"))
        self.pe("(define (compose f g) (lambda (x) (f (g x))))")
        self.pe("(define (square x) (* x x))")
        self.pe("(define (double x) (* x 2))")
        self.assertEquals(64, self.pe("(apply (compose square double) '(4))"))


    def testTooManyAndTooFew(self):
        self.pe("(define (square x) (* x x))")
        self.assertRaises(SchemeError, self.pe, "(square)")
        self.assertRaises(SchemeError, self.pe, "(square 1 2)")
        self.pe("(define (f x . y) (cons x y))")
        self.assertRaises(SchemeError, self.pe, "(f)")


    def testAnotherQuasiquote(self):
        self.pe("(define x 42)")
        self.assertEquals(42, self.pe("x"))
        self.assertEquals(pair.list
                          (Symbol("x"),
                           pair.list(Symbol("quote"),
                                           pair.list(Symbol("x"), 42))),
                          self.pe("`(x '(x ,x))"))

    def testLet(self):
        self.assertEquals(5,
                          self.pe("""(let ((x 3)
                                           (y 4)
                                           (z 2))
                                        (+ x z))"""))

    def testAssignment(self):
        self.pe("(define x 0)")
        self.pe("""(define (inc*2)
                      (set! x (+ x 1))
                      (set! x (+ x 1)))""")
        self.pe("(inc*2)")
        self.assertEquals(2, self.pe("x"))


    def defineRecursiveFunctions(self):
        self.pe("""(define (factorial x)
                      (if (= x 0) 1 (* x (factorial (- x 1)))))""")

        self.pe("""(define (even? x) (if (= x 0) #t (odd? (- x 1))))""")
        self.pe("""(define (odd? x) (if (= x 0) #f (even? (- x 1))))""")

        
    def testRecursion(self):
        self.defineRecursiveFunctions()
        self.assertEquals(6, self.pe("(factorial 3)"))
        self.assertEquals(symbol.true, self.pe("(even? 4)"))


    def testTailRecursiveEval(self):
        """This is a test to see that eval() doesn't break, even on
        deeply recursive functions.  Basically, this is a test of the
        trampoline."""
        self.defineRecursiveFunctions()
        def myFactorial(n):
            return reduce(lambda x, y: x*y, range(1,n+1), 1)
        self.assertEquals(myFactorial(50), self.pe("(factorial 50)"))
        self.assertEquals(symbol.true, self.pe("(even? 42)"))



    def testArithmetic(self):
        self.assertEquals(3, self.pe("(+ 1 2)"))
        self.assertEquals(7, self.pe("(- 8 1)"))
        self.assertEquals(42, self.pe("(* 2 21)"))
        self.assertEquals(2, self.pe("(/ 4 2)"))



    def testIf(self):
        self.assertEquals(Symbol("danny"), self.pe("(if 42 'danny 'yoo)"))
        self.assertEquals(Symbol("yoo"), self.pe("(if #f 'danny 'yoo)"))



    def testCond(self):
        self.assertEquals(Symbol("ok"), self.pe(
            """(define (even? x) (= (remainder x 2) 0))"""))
        self.assertEquals(Symbol("ok"), self.pe(
            """(define (odd? x) (not (even? x)))"""))
        self.assertEquals(Symbol("ok"), self.pe(
            """(cond ((even? 3) "not-ok")
               ((odd? 2) "still-not-ok")
               (#t 'ok))"""))




    def testListPrimitives(self):
        self.assertEquals(symbol.true, self.pe("(list? '(1 2 3))"))
        self.assertEquals(symbol.false, self.pe("(list? (cons 1 (cons 2 3)))"))


    def testStringEvaluating(self):
        self.assertEquals("hello",
                          self.pe("\"hello\""))


    def testLambda(self):
        self.pe("""(define square
                        (lambda (x) (* x x)))""")
        self.assertEquals(49, self.pe("(square 7)"))


    def testWithYOperator(self):
        """A stress test with the infamous Y operator.  This one comes from
        Taming the Y Operator, by Guillermo Juan Rozas."""
        self.pe("""(define (y f)
                      ((lambda (g) (g g))
                       (lambda (x)
                         (f (lambda ()
                              (x x))))))""")
        self.pe("""(define factorial
                      (y (lambda (self)
                            (lambda (n)
                               (if (= n 0)
                                   1
                                   (* n ((self) (- n 1))))))))""")
        self.assertEquals(3628800, self.pe("(factorial 10)"))


    def testGensymConstruction(self):
        self.pe("""(define (make-counter prefix)
                      (let ((i 0))
                         (lambda ()
                           (begin
                              (set! i (+ i 1))
                              (string->symbol
                                 (string-append (symbol->string prefix)
                                                (number->string i)))))))""")
        self.pe("""(define gensym1 (make-counter 'g))""")
        self.pe("""(define gensym2 (make-counter 'h))""")
        self.assertEquals(Symbol("g1"), self.pe("(gensym1)"))
        self.assertEquals(Symbol("g2"), self.pe("(gensym1)"))
        self.assertEquals(Symbol("g3"), self.pe("(gensym1)"))
        self.assertEquals(Symbol("h1"), self.pe("(gensym2)"))

                                 
    def testAndAndOr(self):
        self.pe("""(define (how-tall? height)
                      (cond ((or (<= height 0) (>= height 12)) 'impossible)
                            ((and (< 0 height) (< height 3)) 'midget)
                            ((and (<= 3 height) (< height 6.5)) 'medium)
                            ((and (<= 6.5 height) (< height 12)) 'giant)))""")
        
        ## according to http://usedwigs.com/lists.html
        self.pe("(define mini-me (+ 2 (/ 8.0 12)))")

        ## according to http://www.geocities.com/Colosseum/3522/facts.htm
        self.pe("(define shaq (+ 7 (/ 1.0 12)))")

        ## according to http://animatedtv.about.com/cs/faqs/a/simpsonbios.htm
        self.pe("(define homer 6)")  

        self.assertEquals(Symbol("giant"), self.pe("(how-tall? shaq)"))
        self.assertEquals(Symbol("midget"), self.pe("(how-tall? mini-me)"))
        self.assertEquals(Symbol("medium"), self.pe("(how-tall? homer)"))
        self.assertEquals(Symbol("impossible"), self.pe("(how-tall? 0)"))
        self.assertEquals(Symbol("impossible"), self.pe("(how-tall? 100)"))        


    def testSetCarCdr(self):
        self.pe("(define x '(hello world))")
        self.pe("(define y x)")
        self.pe("(define z (cons 'hi 'earth))")
        self.pe("(set-car! x 'hi)")
        self.pe("(set-cdr! y 'earth)")
        self.assertEquals(pair.cons(Symbol("hi"), Symbol("earth")),
                          self.pe("x"))
        self.assert_(self.pe("(eq? x y)"))
        self.assert_(self.pe("(not (eq? x z))"))
        self.assert_(self.pe("(equal? x z)"))


    def defineCons(self):
        self.pe("""(define cons
                      (lambda (x y)
                         (lambda (dispatch)
                            (if (= dispatch 'car)
                                x
                                y))))""")

    def testApplication(self):
        self.defineCons()
        self.assertEquals(Symbol("ok"),
                          self.pe("(define myname (cons 'danny 'yoo))"))
        self.assertEquals(Symbol("danny"),
                          self.pe("(myname 'car)"))
        self.assertEquals(Symbol("yoo"),
                          self.pe("(myname 'cdr)"))





class ExtendedSchemeTests:
    """Another set of tests that explore some of the more esoteric
    features.  Assumes that there's a pe() method that does parse/eval."""
    def testEvalCallcc(self):
        self.assertEquals(42, self.pe("(call/cc (lambda (x) 42))"))
        self.assertEquals(42, self.pe("(call/cc (lambda (x) (x 42)))"))
        self.assertEquals(15, self.pe("(+ 1 2 3 4 (call/cc (lambda (x) 5)))"))
        self.assertEquals(5, self.pe("(call/cc (lambda (x) (+ 1 2 3 4 (x 5))))"))
        self.pe("""
            (define (member x l)
               (call/cc (lambda (exit)
                            (cond ((null? l) (exit #f))
                                  ((equal? x (car l)) l)
                                  (#t (member x (cdr l)))))))""")
        self.assertEquals(self.pe("(list 4 5)"),
                          self.pe("(member 4 (list 1 2 3 4 5))"))
    

    def testLoad(self):
        ## A small test of the stack module in the 't' test directory.
        self.pe('(load "t/stack.scm")')
        self.pe("(define s (make-stack))")
        self.pe("(push s 42)")
        self.pe("(push s 'foobar)")
        self.pe("(push s (lambda (x) x))")
        self.assertEquals(Symbol("muhaha"),
                          self.pe("((pop s) 'muhaha)"))
        self.assertEquals(Symbol("foobar"), self.pe("(pop s)"))
        self.assertEquals(42, self.pe("(pop s)"))
        self.assertRaises(SchemeError, self.pe, "(pop s)")


    def testDir(self):
        self.assertEquals(pair.list(),
                          self.pe("((lambda () (dir)))"))
        self.assertEquals(pair.list(Symbol("x")),
                          self.pe("((lambda (x) (dir)) 42)"))
    

    def testSchemeEval(self):
        self.assertEquals(42, self.pe("(+ 2 (eval '(+ 30 10)))"))
        self.assertEquals(42, self.pe("(+ 2 (eval '(+ 30 ((lambda () 10)))))"))
        self.assertEquals(Symbol("hello"),
                          self.pe("((eval '(lambda () 'hello)))"))

        ## Quick test to see that EVAL'ed expressions are also
        ## expanded, since AND is derived.
        self.assertEquals(Symbol("foo"),
                          self.pe("(eval '(AND 'this 'is 'a 'foo))"))


    def testParse(self):
        self.assertEquals(parser.parse("(this is ( a test))"),
                          self.pe('(parse "(this is ( a test))")'))




class BasicSchemeTestCase(RegularInterpreterMixin,
                          BasicSchemeTests, unittest.TestCase):
    pass


class ExtendedSchemeTestCase(RegularInterpreterMixin,
                             ExtendedSchemeTests, unittest.TestCase):
    pass


if __name__ == '__main__':
    unittest.main()



"""Package to hold pyscheme together.

We also keep a few utility functions in here.
"""


__license__ = "MIT License"

import scheme
import pair
import parser
import symbol



def make_interpreter():
    """Simple utility function to make an pyscheme.scheme.Interpreter.
    Defaults to AnalyzingInterpreter for now."""
    return scheme.AnalyzingInterpreter()


def parse(s):
    """Parses string 's' into list structure."""
    return parser.parse(s)



"""A simple "symbol" class for Python.  Symbols are like strings,
except that symbols should only be comparable to other symbols.  This
module is designed so that one should be able to do a "from symbol
import *" on this.

To create a new Symbol, use the Symbol() function, like this:

    hello = Symbol("hello")

All symbols are interned so that comparison using 'is' will work.  For
example,

    Symbol("hello") is Symbol("h" + "ello")

should work.

By default, all symbols are case insensitive, but if the module
variable CASE_SENSITIVE is set to a true value, symbols will become
case sensitive.

Also, we use the UserString to make sure that symbols are of a
different "class" than Strings.  isSymbol() depends on this
representation, so be careful about changing it.

"""

__license__ = "MIT License"


from UserString import UserString as __UserString
import weakref

__all__ = ['Symbol', 'isSymbol']


"""By default, symbols are case-insensitive."""
CASE_SENSITIVE = 0


class __Symbol(__UserString):
    """Support class for symbols."""
    def __eq__(self, other):
        """Comparison should only be possible between symbols.  Since all
        Symbols are interned, this is equivalent to an 'is' check."""
        return self is other


"""A global dictionary that contains all known symbols.
__interned_symbols: strings -> symbols

The dictionary uses weak references to reduce the chance of symbol abuse.
"""
__interned_symbols = weakref.WeakValueDictionary({})


def Symbol(s):
    """Generates a new Symbol that we guarantee can be compared in
    constant time to any other symbol."""
    global __interned_symbols

    ## Defensive programming measure:
    assert not isinstance(s, __Symbol), ("%s already a symbol!" % s)

    if not CASE_SENSITIVE: s = s.lower()
    if not __interned_symbols.has_key(s):
        ## Subtle note: we have to use a local variable here to store the newSymbol.
        ## If we had tried something shorter like:
        ##
        ##     __interned_symbol[s] = __Symbol(s)
        ##
        ## then the new Symbol will immediately evaporate since there
        ## aren't any hard references!  Weak references can be tricky.
        newSymbol = __Symbol(s)
        __interned_symbols[s] = newSymbol
    return __interned_symbols[s]




"""Here are definitions of symbols that we should know about."""
false = Symbol("#f")
true = Symbol("#t")
__empty_symbol = Symbol("")


def isSymbol(x):
    """Returns True if x is already a symbol."""
    return type(x) == type(__empty_symbol)



def makeUniqueTemporary(_counter = [0]):
    """Constructs a symbol that does not collide with any other
    symbol.  I'll use this to help with the macro-expansion.

    NOTE/FIXME: we do this by making an "illegal" symbol which starts
    with a number and contains a space.  The parser module doesn't
    allow such symbols to exist... although it's possible to subvert this
    by calling STRING->SYMBOL.  So this mechanism is not perfect.
    """
    while ('%d*** temporary' % _counter[0]) in __interned_symbols:
        _counter[0] += 1
    return Symbol('%d*** temporary' % _counter[0])



"""Environment stuff.  Exercise 4.11 asks us to develop an alternative
environment implementation.  This is an implementation of environments
that uses a chained list of Python dictionaries.
"""

__license__ = "MIT License"

from symbol import Symbol
from error import SchemeError
import pair

def enclosingEnvironment(env):
    return env[1:]

def firstFrame(env):
    return env[0]


THE_EMPTY_ENVIRONMENT = []


def makeFrame(var_pairs, val_pairs):
    """Note: here I diverge from SICP's implementation: instead of
    using a cons, I use a Python list.  The selectors frameVariables()
    and frameValues() reflect this."""
    car, cdr, isNull = pair.car, pair.cdr, pair.isNull
    frame = {}
    while not isNull(var_pairs):
        if pair.isPair(var_pairs):
            if not isNull(val_pairs):
                frame[car(var_pairs)] = car(val_pairs)
                var_pairs, val_pairs = cdr(var_pairs), cdr(val_pairs)
            else:
                raise SchemeError, "Too few arguments supplied"
        else:
            frame[var_pairs] = val_pairs
            var_pairs, val_pairs = pair.NIL, pair.NIL
    if isNull(val_pairs):
        return frame
    else:
        raise SchemeError, "Too many arguments supplied"




def addBindingToFrame(var, val, frame):
    frame[var] = val


def extendEnvironment(var_pairs, val_pairs, base_env):
    """Extends an environment with a new set of bindings.  Extended to
    support Scheme's dotted notation for varargs."""
##     if Symbol('.') in vars:
##         vararg_start = vars.index(Symbol('.'))
##         return [makeFrame(vars[:vararg_start] + [vars[vararg_start+1]],
##                           vals[:vararg_start] + [vals[vararg_start:]])
##                 ] + base_env
##     elif len(vars) == len(vals):
    return [makeFrame(var_pairs, val_pairs)] + base_env
## FIXME: add length check!
##     elif len(vars) < len(vals):
##         raise SchemeError, \
##               "Too many arguments supplied %s %s" % (vars, vals)
##     raise SchemeError, "Too few arguments supplied %s %s" % (vars, vals)




def lookupVariableValue(var, env):
    while 1:
        if env == THE_EMPTY_ENVIRONMENT:
            raise SchemeError, "Unbound variable " + var
        frame = firstFrame(env)
        if frame.has_key(var):
            return frame[var]
        env = enclosingEnvironment(env)
    

def setVariableValue(var, val, env):
    while 1:
        if env == THE_EMPTY_ENVIRONMENT:
            raise SchemeError, "Unbound variable -- SET! " + var
        frame = firstFrame(env)
        if frame.has_key(var):
            frame[var] = val
            return
        env = enclosingEnvironment(env)


def defineVariable(var, val, env):
    frame = firstFrame(env)
    if frame.has_key(var):
        frame[var] = val
        return
    addBindingToFrame(var, val, frame)





import unittest

import sys

import analyzer
import parser
import test_scheme
from environment import THE_EMPTY_ENVIRONMENT, extendEnvironment, defineVariable
from symbol import Symbol
import pair
from error import SchemeError
import scheme


class AnalyzerEvaluatorMixin:
    def setUp(self):
         ## sets the recursion limit to something fairly small to test
         ## out the the tail recursion.
        self.old_recursion_limit = sys.getrecursionlimit()
        sys.setrecursionlimit(55)
        self.interp = scheme.AnalyzingInterpreter()

    def tearDown(self):
        sys.setrecursionlimit(self.old_recursion_limit)


    def pe(self, s):
        """Parse and evaluate."""
        return self.interp.eval(parser.parse(s))



class AnalyzingBasicSchemeTestCase(
    AnalyzerEvaluatorMixin, test_scheme.BasicSchemeTests, unittest.TestCase):
    pass


class AnalyzingExtendedSchemeTestCase(
    AnalyzerEvaluatorMixin, test_scheme.ExtendedSchemeTests, unittest.TestCase):
    pass



    
class IndependentAnalyzerTests(unittest.TestCase):
    def setUp(self):
        ## We set up a VERY minimal environment here for some tests.
        ## We also set the recursion limit to something dreadful to see that
        ## call/cc is doing the right thing.
        ##

        ## Note: these tests directly work with analyzer.eval, and not
        ## through the nicer scheme.AnalyzingInterpreter interface.

        self.env = extendEnvironment(pair.list(Symbol('pi')),
                                     pair.list(3.1415926),
                                     THE_EMPTY_ENVIRONMENT)
        defineVariable(Symbol("#t"), Symbol("#t"), self.env)
        defineVariable(Symbol("#f"), Symbol("#f"), self.env)
        self.old_recursion_limit = sys.getrecursionlimit()
        sys.setrecursionlimit(100)


    def tearDown(self):
        self.env = None
        sys.setrecursionlimit(self.old_recursion_limit)


    def pe(self, s):
        """Parse and evaluate."""
        return analyzer.eval(parser.parse(s),
                             self.env)




    def testSelfEvaluating(self):
        self.assertEquals(42, self.pe("42"))
        self.assertEquals(pair.list(), self.pe("()"))


    def testVariable(self):
        self.assertEquals(3.1415926, self.pe("pi"))
        self.assertRaises(SchemeError, self.pe, "nonexistantvalue")


    def testQuotation(self):
        self.assertEquals(Symbol("foobar"), self.pe("'foobar"))
        self.assertEquals(Symbol("foobar"), self.pe("'foobar"))
        self.assertEquals(pair.list(1, 2, 3), self.pe("'(1 2 3)"))


    def testAssignment(self):
        self.assertEquals(Symbol("ok"),
                          self.pe("(set! pi 'three-point-one-four)"))
        self.assertEquals(Symbol("three-point-one-four"),
                          self.pe("pi"))
        self.assertRaises(SchemeError, self.pe, "(set! nonexistantvalue 42)")



    def testDefinition(self):
        self.assertEquals(Symbol("ok"),
                          self.pe("(define name 'danny)"))
        self.assertEquals(Symbol("danny"), self.pe("name"))


    def testIf(self):
        self.assertEquals(Symbol("ok"), self.pe("(if #t 'ok 'not-ok)"))
        self.assertEquals(Symbol("ok"), self.pe("(if #f 'not-ok 'ok)"))


    def testLambda(self):
        ## Can't test much yet without application.
        self.pe("(lambda (x) 'hello 'world)")

        
    def testBegin(self):
        self.assertEquals(Symbol("hello"),
                          self.pe("(begin 'hello)"))
        self.assertEquals(Symbol("world"),
                          self.pe("(begin 'hello '(this is a test) 'world)"))
        self.assertRaises(SchemeError, self.pe, "(begin)")

        
    def testSimpleApplication(self):
        self.assertEquals(Symbol("hello"),
                          self.pe("((lambda (x) x) 'hello)"))
        self.assertEquals(Symbol("hi"),
                          self.pe("((lambda (x y) y) '(this is a test) 'hi)"))



if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

"""A translation of a Scheme interpreter in Python, modeling the
metacircular evaluator in Ch. 4 of The Structure and Interpretation of
Computer Programs.  (http://www-mitpress.mit.edu/sicp/)

AUTHOR:

    Danny Yoo (dyoo@hkn.eecs.berkeley.edu)

SYNOPSIS:

    It's Scheme.  *grin*  At least, quite a bit of Scheme.

    Scheme is a variant of Lisp, and it's used in many introductory CS
    courses.  It is pretty powerful despite its deceptive size.  For
    more information on Scheme, visit the Scheme home page:

        http://www.swiss.ai.mit.edu/projects/scheme


OLD BUGS:

    One major deficiency of this implementation had in the past was
    the lack of tail call elimination.  However, I've gotten this to
    work by using trampolined style.  This involved rewriting the
    interpreter into CPS form, and trampolining the resulting tail
    calls.  The resulting interpreter looks quite a bit uglier, but at
    least it shouldn't stack overflow easily.

BUGS:

    Not all of R5RS is implemented, and probably will never be until I
    finish reading and understanding it.


MISCELLANEOUS:

    This source code is for the public domain; I'm writing it just for
    the fun of it.  This is certainly not meant to be efficient or
    even useful.  *grin*

    My future goals is to make this usable enough to run something
    like the explicit-register simulator and compiler in SICP Ch. 5.
    I'm learning more about the advanced features in scheme
    (continuations, hygenic macros), so I'll probably use this as a
    future testbed to make sure I understand those ideas.
"""

__license__ = "MIT License"


import sys
import optparse
import symbol
import parser
import evaluator
import builtins
import analyzer
import prompt
import error
import expressions
import environment
import pogo
import pair
import expander

######################################################################


class Interpreter:
    """Meant to be subclassed.
    Subclasses should define _eval and _env in their constructors.

    _eval should be a callable that takes an expression and
    environment, and returns a value.

    _env should be an environment.

    """

    def __init__(self):
        self._eval = None
        self._env = None
        self._expander = expander.Expander()
        expander.install_core_handlers(self._expander)


    def eval(self, exp):
        """Evaluates an expression exp."""
        return self.get_evaluator()(self.get_expander().expand(exp),
                                    self.get_environment())


    def install_function(self, name, function):
        """Installs a Python function into the toplevel environment."""
        builtins.installPythonFunction(name, function, self.get_environment())


    def repl(self):
        """A thin method around the repl function."""
        return repl(self)

        
    def get_evaluator(self):
        """Returns the evaluation function."""
        return self._eval


    def get_environment(self):
        """Returns the toplevel environment frame."""
        return self._env

    def get_expander(self):
        """Returns the syntax expander."""
        return self._expander


    def install_special_builtins(self):
        """Adds some unusual builtin cases that involve the interpreter
        itself."""
        for (name, function) in self.get_special_builtins():
            builtins.installPythonFunction(name,
                                           function,
                                           self.get_environment(),
                                           wrapDefaults=False)

    def get_special_builtins(self):
        """Returns a list of 2-tuples: name of the special builtin, and
        the function value of the builtin."""
        return ()





class RegularInterpreter(Interpreter):
    """This interpreter evaluates expressions in a way similar to
    the metacircular evaluator described by SICP.  The only significant
    difference is the internal use of trampolining style plus CPS."""
    def __init__(self):
        Interpreter.__init__(self)
        self._eval = evaluator.eval
        self._env = builtins.setupEnvironment()
        self.install_special_builtins()


    def get_special_builtins(self):
        return (('load', self.schemeLoad),
                ('call/cc', self.schemeCallcc),
                ('call-with-current-continuation', self.schemeCallcc),
                ('eval', self.schemeEval),
                ('apply', self.schemeApply),
                ('dir', self.schemeDir))


    def schemeDir(self, cont, env, args):
        """A quick function to inspect the first frame."""
        if len(args) != 0:
            raise TypeError, ("dir expected at most 0 arguments, got %s" % len(args))
        names = environment.firstFrame(env).keys()
        names.sort()
        return pogo.bounce(cont, pair.list(*names))


    def schemeLoad(self, cont, env, args):
        """Special primitive: implements LOAD."""
        symbolicFilename = str(args[0])
        try:
            f = open(symbolicFilename)
            try:
                text = "(begin \n%s\n 'ok)" % f.read()
            finally:
                f.close()
            expandedExp = self.get_expander().expand(parser.parse(text))
            return evaluator.teval(expandedExp, env, cont)
        except IOError, e:
            raise error.SchemeError, "LOAD error -- %s" % str(e)
        except TypeError, e:
            raise error.SchemeError, "LOAD error -- argument must be a string"


    def schemeApply(self, cont, env, args):
        return evaluator.apply(args[0], args[1], env, cont)


    def schemeEval(self, cont, env, args):
        expandedExp = self.get_expander().expand(args[0])
        return evaluator.teval(expandedExp, env, cont)


    def schemeCallcc(self, cont, env, args):
        lambdaBody = args[0]
        def c2(val):
            return pogo.bounce(cont, val)
        cont_procedure = expressions.makeContinuationProcedure(c2)
        return evaluator.apply(lambdaBody, pair.list(cont_procedure),
                               env, cont)



######################################################################



class AnalyzingInterpreter(Interpreter):
    """This version of the interpreter will do syntax analysis on its
    expressions before it executes them.  Should be a win on speed."""
    def __init__(self):
        Interpreter.__init__(self)
        self._eval = analyzer.eval
        self._env = builtins.setupEnvironment()                       
        self.install_special_builtins()


    def get_special_builtins(self):
        return ( ('load', self.schemeLoad),
                 ('dir', self.schemeDir),
                 ('call-with-current-continuation', self.schemeCallcc),
                 ('call/cc', self.schemeCallcc),
                 ('apply', self.schemeApply),
                 ('eval', self.schemeEval),
                 )

    def schemeDir(self, cont, env, args):
        """A quick function to inspect the first frame."""
        if len(args) != 0:
            raise TypeError, ("dir expected at most 0 arguments, got %s" % len(args))
        names = environment.firstFrame(env).keys()
        names.sort()
        return pogo.bounce(cont, pair.list(*names))


    def schemeLoad(self, cont, env, args):
        """Special primitive: implements LOAD."""
        symbolicFilename = str(args[0])
        try:
            f = open(symbolicFilename)
            try:
                text = "(begin \n%s\n 'ok)" % f.read()
            finally:
                f.close()
            expandedExp = self.get_expander().expand(parser.parse(text))
            analyzedExp = analyzer.analyze(expandedExp)
            return analyzer.texec(analyzedExp, env, cont)
        except IOError, e:
            raise error.SchemeError, "LOAD error -- %s" % str(e)
        except TypeError, e:
            raise error.SchemeError, "LOAD error -- argument must be a string"


    def schemeApply(self, cont, env, args):
        return analyzer.apply(args[0],
                              args[1],
                              env, cont)


    def schemeEval(self, cont, env, args):
        expandedExp = self.get_expander().expand(args[0])
        analyzedExp = analyzer.analyze(expandedExp)
        return analyzer.texec(analyzedExp, env, cont)


    def schemeCallcc(self, cont, env, args):
        analyzedLambdaBody = args[0]
        def c2(val):
            return pogo.bounce(cont, val)
        cont_procedure = expressions.makeContinuationProcedure(c2)
        return analyzer.apply(analyzedLambdaBody,
                              pair.list(cont_procedure),
                              env,
                              cont)



######################################################################


class MinimalInterpreter(Interpreter):
    """
    This is just a really minimal interpreter that supports only the
    core forms.  No builtins, no expander, nothing except the
    following:

        self evaluating expressions
        variable references
        QUOTE
        SET!
        DEFINE
        IF
        LAMBDA
        BEGIN
        function application
    
    Furthermore, each evaluation generates a whole new environment, so
    no state should be saved between calls to eval.

    This is a demonstration of a restricted Scheme.

    In the future, we may even want to modify EVAL so that it goes
    through a version of pogo.pogo that bounds the number of bounces
    allowed before getting an answer.  So we can do some interesting
    things here...
    """

    def __init__(self):
        Interpreter.__init__(self)

    def get_evaluator(self):
        return analyzer.eval

    def get_environment(self):
        return environment.extendEnvironment(pair.list(),
                                             pair.list(),
                                             environment.THE_EMPTY_ENVIRONMENT)
    def get_expander(self):
        class SillyExpander:
            def expand(self, expr): return expr
        return SillyExpander()



######################################################################


def repl(interp=AnalyzingInterpreter()):
    """A fairly standard 'read->eval->print' loop."""
    print 'Welcome to PyScheme!  Type: (QUIT) to quit.\n'
    def promptCallback(s):
        try:
            print expressions.toString(interp.eval(parser.parse(s)))
        except (error.SchemeError, parser.ParserError), e:
            print "ERROR (%s): %s" % (e.__class__, e)

    p = prompt.Prompt(name="PyScheme", callback=promptCallback,
                      quit_str=None)
    try:
        p.promptLoop()
    except SystemExit:
        pass ## Get out smoothly.
            

def isNonInteractive(args):
    """Returns True if we want to run in non-interactive mode."""
    return len(args) >= 1

def runNonInteractive(interp, args):
    """Executes the interpreter on args[0]."""
    interp.eval(pair.list(symbol.Symbol("load"),
                                args[0]))


def chooseInterpreter(options):
    """Choose which interpreter we should use."""
    if options.interpreter_type == 'regular':
        return RegularInterpreter()
    else:
        return AnalyzingInterpreter()


def configureOptionParser():
    optparser = optparse.OptionParser()
    optparser.add_option("--regular",
                      help="invoke regular interpreter",
                      action="store_const",
                      const="regular",
                      dest="interpreter_type",
                      )
    optparser.add_option("--analyzer",
                      help="invoke analyzing interpreter [default]",
                      action="store_const",
                      const="analyzer",
                      dest="interpreter_type",
                      )
    return optparser


def main():
    """Main driver."""
    optparser = configureOptionParser()
    options, args = optparser.parse_args()
    interp = chooseInterpreter(options)

    if isNonInteractive(args):
        runNonInteractive(interp, args)
    else:
        repl(interp)



if __name__ == '__main__':
    main()




# Python imports
import os
import sys
import copy
import base64
import pprint
import struct
import traceback

if sys.platform == "darwin":
	import pickle as pickle
else:
	import cPickle as pickle
from datetime import datetime

def df(time):
	if type(time) in (float, int, long):
		return datetime.utcfromtimestamp(time).strftime('%c')
	elif type(time) is datetime:
		return time.strftime('%c')
	else:
		raise TypeError("Unable to output this type...")

try:
	set()
except NameError:
	from sets import Set as set

# Other library imports
from tp.netlib import Connection, failed, constants, objects
from tp.netlib.objects import Header, Description, OrderDescs, DynamicBaseOrder

# Local imports
# FIXME: Should I think about merging the ChangeList and ChangeDict?
from ChangeDict import ChangeDict
from ChangeList import ChangeList, ChangeNode

from threads import Event
from threadcheck import thread_checker, thread_safe

__CACHE = None
class Cache(object):
	"""\
	This is the a cache of the data downloaded from the network. 

	It can be pickled and restored at a later date to preserve the data accross application runs.
	"""
	__metaclass__ = thread_checker

	version = 5

	class CacheEvent(Event):
		"""\
		Raised when the game cache is made dirty. Contains a reference to what was updated.
		"""
		def __init__(self, what, action, id, *args, **kw):
			Event.__init__(self)

			if what in Cache.readonly:
				raise ValueError("Can not change that!")
			elif not what in Cache.readwrite:
				raise ValueError("Invalid value (%s) for what" % (what,))
			else:
				self.what = what

			if what in Cache.compound:
				if not action in Cache.actions_compound:
					raise ValueError("Invalid action (%s)" % (action,))
			else:
				if not action in Cache.actions:
					raise ValueError("Invalid action (%s)" % (action,))
			self.action = action

			self.id = id

			args = list(args)
			if what in Cache.compound:
				if len(args) == 2:
					self.node = args.pop(0)
				elif kw.has_key('node'):
					self.node = kw['node']
				elif kw.has_key('nodes'):
					if not action is "remove":
						raise ValueError("Slots is only valid with a remove action")
					self.node  = None
					self.nodes = kw['nodes']
				else:
					raise TypeError("A node is required for compound types.")
				if not hasattr(self, "nodes"):
					self.nodes = [self.node]

			# Do a type check for the nodes
			for node in self.nodes:
				if (not node is None) and not isinstance(node, ChangeNode):
					raise TypeError("Nodes must be of type ChangeNode not %s (%r)" % (type(node), node))

				assert node.inlist()

			if len(args) == 1:
				self.change = args.pop(0)
			elif kw.has_key('change'):
				self.change = kw['change']
			elif action is "remove":
				pass
			else:
				raise TypeError("The actual change needs to be added.")

		def __str__(self):
			if not self.what:
				return "<%s full-update>" % (self.__class__.__name__,)
			elif hasattr(self, 'node'):
				if self.node is None:
					return "<%s %s %s id=%i nodes=%r>" % (self.__class__.__name__, self.what, self.action, self.id, self.nodes)
				else:
					return "<%s %s %s id=%i node=%r>" % (self.__class__.__name__, self.what, self.action, self.id, self.node)
			else:
				return "<%s %s %s id=%i>" % (self.__class__.__name__, self.what, self.action, self.id)

		__repr__ = __str__

	class CacheDirtyEvent(CacheEvent):
		"""\
		Raised when the game cache is made dirty. Contains a reference to what was updated.
		"""
		pass

	class CacheUpdateEvent(CacheEvent):
		"""\
		Raised when the game cache is changed. Contains a reference to what was updated. 
		If the what is None a new cache has been created.
		"""
		def __init__(self, what, *args, **kw):
			if what == None:
				Event.__init__(self)
				self.what = None
			else:
				CacheEvent.__init__(self, what, *args, **kw)

	# Read Only things can only be updated via the network
	readonly = ("features", "objects", "orders_probe", "boards", "resources", "components", "properties", "players", "resources")
	# These can be updated via either side
	readwrite = ("orders", "messages", "categories", "designs")
	# How we can update the Cache
	actions = ("create", "remove", "change")
	actions_compound = ("create before", "create after", "remove", "change")
	compound = ("orders", "messages")

	@staticmethod
	def key(server, username):
		key = server

		p = ['tp://', 'tps://', 'http://', 'https://']
		found = False
		for p in p:
			if key.startswith(p):
				found = True
				break
		if not found:
			key = 'tp://' + key
		if key.find('@') == -1:
			p, s = key.split('//', 1)
			key = "%s//%s@%s" % (p, username, s)
		return key

	@staticmethod
	def configkey(key):
		key = base64.encodestring(key)[:-2]
		return key

	@staticmethod
	def configdir():
		dirs = [("APPDATA", "Thousand Parsec"), ("HOME", ".tp"), (".", "var")]
		for base, extra in dirs:
			if base in os.environ:
				base = os.environ[base]
				break
			elif base != ".":
				continue

		return os.path.join(base, extra)

	def __init__(self, key, configdir=None, new=False):
		"""\
		It is important that key constructed the following way,

		protocol://username@server:port/

		Everything must be there, even if the port is the default.
		"""
		if configdir == None:
			configdir = Cache.configdir()

		if not os.path.exists(configdir):
			os.mkdir(configdir)

		key = Cache.configkey(key)

		self.file = os.path.join(configdir, "cache.%s" % (key,))
		if os.path.exists(self.file) and not new:
			# Load the previously cached status
			print "Loading previous saved data (from %s)." % (self.file,)
			try:
				self.load()
				return
			except (IOError, EOFError, KeyError, pickle.PickleError), e:
				print e
				traceback.print_exc()
				print "Unable to load the data, saved cache must be corrupt."
		print "Creating the Cache fresh (%s)." % (self.file,)
		self.new()

	def new(self):
		# Features
		self.features		= []

		# The object stuff
		self.objects		= ChangeDict()
		self.orders			= ChangeDict()
		self.orders_probe	= ChangeDict()

		# The message boards
		self.boards			= ChangeDict()
		self.messages		= ChangeDict()

		# Design stuff
		self.categories		= ChangeDict()
		self.designs		= ChangeDict()
		self.components		= ChangeDict()
		self.properties		= ChangeDict()

		self.players		= ChangeDict()
		self.resources		= ChangeDict()

	@thread_safe # FIXME: This probably isn't thread safe!
	def apply(self, *args, **kw):
		"""\
		Given a CacheDirtyEvent, it sets up the changes...
		"""
		evt = Cache.CacheDirtyEvent(*args, **kw)

		if not evt.what in Cache.compound:
			return evt
		else:
			d = getattr(self, evt.what)[evt.id]

			if evt.action == "remove":
				for node in evt.nodes:
					node.AddState("removing")
				evt.change = None

			elif evt.action == "change":
				assert len(evt.nodes) == 1
				assert not evt.node.LastState in ("removed", "removing")

				evt.node.AddState("updating", evt.change)
				evt.change = evt.node

			elif evt.action.startswith("create"):
				assert len(evt.nodes) == 1

				# Create the new node
				newnode = ChangeNode(None)
				newnode.AddState("creating", evt.change)

				# Insert the node
				if evt.action == "create after":
					d.insert_after(evt.node,  newnode)
				elif evt.action == "create before":
					d.insert_before(evt.node, newnode)
				else:
					assert False, "Unknown action!"

				# Set the new node as the change
				evt.change = newnode
			else:
				assert False, "Unknown action!"

			return evt

	def commit(self, evt):
		"""\
		Given a CacheDirtyEvent it applies the changes to the cache.
		It should be called after the changes have been confimed by the server.

		It then mutates the event into a CacheUpdateEvent.
		"""
		if not isinstance(evt, self.CacheDirtyEvent):
			raise TypeError("I can only accept CacheDirtyEvents")

		if not evt.what in Cache.compound:
			if evt.action == "create" or evt.action == "change":
				getattr(self, evt.what)[evt.id] = (-1, evt.change)
			elif evt.action == "remove":
				del getattr(self, evt.what)[evt.id]

		else:
			d = getattr(self, evt.what)[evt.id]

			if evt.action.startswith("create") or evt.action == "change":
				node = evt.change

				assert node.CurrentState in ("creating", "updating"), "Current state (%s) doesn't match action %s" % (node.CurrentState, evt.action)

				node.PopState()

			elif evt.action == "remove":
				for node in evt.nodes:
					assert node.CurrentState == "removing"

					del d[node.id]
					node.PopState()
			else:
				raise SystemError("Unknown node state!")


		evt.__class__ = self.CacheUpdateEvent

	def load(self):
		"""\
		"""
		f = open(self.file, 'rb')

		# Read in the version number
		v, = struct.unpack('!I', f.read(4))
		if v != self.version:
			raise IOError("The cache is not of this version! (It's version %s)" % (v,))

		orderdescs, = struct.unpack('!I', f.read(4))

		# Now load the order cache
		for i in xrange(0, orderdescs):
			d = f.read(Header.size)
			p = Header.fromstr(d)

			d = f.read(p.length)
			p.__process__(d)

			assert isinstance(p, Description)
			p.register()

		# First load the pickle
		d = pickle.load(f)
		if d.has_key('file'):
			del d['file']				# Stop the file being loaded

		for clist in d['orders'].values():
			for node in clist:
				for pending in node._pending:
					pending.__class__ = OrderDescs()[pending.subtype]
				node._what.__class__ = OrderDescs()[node._what.subtype]

		self.__dict__.update(d)

	def save(self):
		"""\
		"""
		# We don't want this filename appearing in the cace
		file = self.file
		del self.file

		# Save the cache
		f = open(file, 'wb')
		f.write(struct.pack('!I', self.version))

		# Save each dynamic order description
		descriptions = OrderDescs()

		f.write(struct.pack('!I', len(descriptions)))
		for orderdesc in descriptions.values():
			f.write(str(orderdesc.packet))

		p = copy.copy(self.__dict__)
		del p['_thread']

		# Stop referencing the dynamic orders
		for clist in p['orders'].values():
			for node in clist:
				for pending in node._pending:
					subtype = pending.subtype
					pending.__class__ = DynamicBaseOrder
					pending.subtype = subtype

				subtype = node._what.subtype
				node._what.__class__ = DynamicBaseOrder
				node._what.subtype = subtype

		pickle.dump(p, f)

		# FIXME: The above copy should not be mutating!
		for clist in p['orders'].values():
			for node in clist:
				for pending in node._pending:
					pending.__class__ = OrderDescs()[pending.subtype]
				node._what.__class__ = OrderDescs()[node._what.subtype]
			
		f.close()

		# Restore the file
		self.file = file

	def update(self, connection, callback):
		"""\
		Updates the cache using the connection.

		The callback function is called in the following way,

		callback(group=<mode>, state=<state>, message=<message>)

		The message string is a human readable message about what is happening.

		Group is the current group of things been updated the possible choices are,
			objects
			orders
			orders_probe
			boards
			messages
			categories
			designs
			components
			properties
			players
			resources

		State is one of the following,
			start		- no more arguments
			todownload	- total, the total number of things to be downloaded
			progress	- some sort of undetermined progress occured
			failure		- some sort of failure when downloading occured
			downloaded	- amount, the number of things which have been downloaded
			finished	- no more arguments
		"""
		c = callback
		# FIXME: We should restart with an empty cache if the following has happened
		#	FIXME: This should compare any read-only attributes and see if they have change
		#	FIXME: This should check the current turn and see if the turn is strange (IE gone back in time)
		#	FIXME: Should check that none of the Order definitions have changed

		# Get the features this server support
#		c("connecting", "todownload", message=_("Looking for supported features..."))
		self.features = connection.features()
#		c("connecting", "finished")

		c("orderdescs", "start", message=_("Getting order descriptions..."))
		c("orderdescs", "progess", message=_("Working out the number of order descriptions to get.."))
		ids = []
		for id, time in connection.get_orderdesc_ids(iter=True):
			if OrderDescs().has_key(id) and hasattr(OrderDescs()[id], "modify_time"):
				if time <= OrderDescs()[id].modify_time:
					continue
			ids.append(id)
		c("orderdescs", "todownload", todownload=len(ids))

		for id in ids:
			desc = connection.get_orderdescs(id=id)[0]
			# Did we download the order description okay?
			if not failed(desc):
				c("orderdescs", "downloaded", amount=1, \
					message=_("Got order description %(order)s (ID: %(id)i) (last modified at %(time)s)...") % {'order': desc._name, 'id': id, 'time': time})
				desc.register()
			else:
				c("orderdescs", "failure",
					message=_("Failed to get order description with ID %(id)i (last modified at %(time)s)...") % {'id': id, 'time': time})

		c("orderdescs", "finished", message=_("Recieved all order descriptions..."))

		# Get all the objects
		#############################################################################
		#############################################################################
		toget = self.__getObjects(connection,   "objects", callback)
		if toget > 0:
			self.__getSubObjects(connection, toget, "objects", "orders", "order_number", callback)
		else:
			c("orders", "finished", message=_("Don't have any orders to get.."))

		toget = self.__getObjects(connection,   "boards", callback)
		if toget > 0:
			self.__getSubObjects(connection, toget, "boards",  "messages", "number", callback)
		else:
			c("messages", "finished", message=_("Don't have any messages to get.."))

		self.__getObjects(connection, "categories", callback)
		self.__getObjects(connection, "designs",    callback)
		self.__getObjects(connection, "components", callback)
		self.__getObjects(connection, "properties", callback)
#		self.__getObjects(connection, "players",    callback)
		self.__getObjects(connection, "resources",  callback)

		c("players", "start", message=_("Getting player objects..."))

		playerids = self.players.keys()
		if len(playerids) > 0:
			playerids.sort()
			i = playerids[-1]+1
		else:
			i = 0

		while True:
			player = connection.get_players(i)

			if failed(player):
				break
			else:
				self.players[i] = player[0]

			c("players", "downloaded", amount=1, \
				message=_("Got player %(player)s (ID: %(id)i)...") % {'player': player[0].name, 'id': player[0].id})

			i += 1

		c("players", "finished", message=_("Received all player objects..."))
		
#		self.players[0] = connection.get_players(0)[0]

	def __getObjects(self, connection, plural_name, callback):
		"""\
		Get a thing which has a container.
		"""
		c = callback
		pn = plural_name

		if pn[-3:] == 'ies':
			sn = pn[:-3]+'y'
		elif pn[-1:] == 's':
			sn = pn[:-1]
		else:
			sn = pn

		def cache(id=None, self=self, pn=pn):
			if id==None:
				return getattr(self, pn)
			else:
				return getattr(self, pn)[id]

		c(pn, "start", message=_("Getting %s...") % pn)

		# Figure out the IDs to download
		c(pn, "progess", message=_("Working out the number of %s to get..") % pn)
		toget = []
		ids = []
		for id, time in getattr(connection, "get_%s_ids" % sn)(iter=True):
			ids.append(id)
			if not cache().has_key(id):
				c(pn, "info", message=_("%(plural_name)s: Getting %(id)s as not cache().has_key(id)") % {'plural_name': pn, 'id': id})
				toget.append(id)
			elif time > cache().times[id]:
				c(pn, "info", message=_("%(plural_name)s: Getting %(id)s (%(name)s) as %(time)s > %(cached_time)s") % {'plural_name': pn, 'id': id, 'name': cache(id).name, 'time': time, 'cached_time': cache().times[id]})
				toget.append(id)
			else:
				c(pn, "info", message=_("%(plural_name)s: Not getting %(id)s (%(name)s) as %(time)s <= %(cached_time)s") % {'plural_name': pn, 'id': id, 'name': cache(id).name, 'time': time, 'cached_time': cache().times[id]})

		# Callback function
		def OnPacket(p, c=c, pn=pn, sn=sn, objects=objects):
			if isinstance(p, getattr(objects, sn.title())):
				c(pn, "downloaded", amount=1, \
					message=_("Got %(singular_name)s %(name)s (ID: %(id)i) (last modified at %(time)s)...") % {'singular_name': sn, 'name': p.name, 'id': p.id, 'time': p.modify_time})

		if len(toget) < 1:
			c(pn, "finished", message=_("No %s to get, skipping...") % pn)
			return 0

		# Download the XXX
		c(pn, "todownload", \
			message=_("Have %(ammount)i %(plural_name)s to get...") % {'ammount': len(toget), 'plural_name': pn}, todownload=len(toget))
		frames = getattr(connection, "get_%s" % pn)(ids=toget, callback=OnPacket)

		if failed(frames):
			raise IOError("Strange error occured, unable to request %s." % pn)

		# Match the results to the associated ids
		for id, frame in zip(toget, frames):
			if not failed(frame):
				if cache().has_key(id):
					c(pn, "info", \
						message=_("%(plural_name)s: Updating %(id)s (%(name)s - %(frame_name)s) with modtime %(time)s") % {'plural_name': pn, 'id': id, 'name': cache(id).name, 'frame_name': frame.name, 'time': frame.modify_time})
				else:
					c(pn, "info", \
						message=_("%(plural_name)s: Updating %(id)s (%(frame_name)s - New!) with modtime %(time)s") % {'plural_name': pn, 'id': id, 'frame_name': frame.name, 'time': frame.modify_time})
				cache()[id] = (frame.modify_time, frame)
			else:
				if cache().has_key(id):
					c(pn, "failure", \
						message=_("Failed to get the %(singular_name)s which was previously called %(name)s.") % {'singular_name': sn, 'name': cache(id).name})
				else:
					c(pn, "failure", \
						message=_("Failed to get the %(singular_name)s with ID %(id)s.") % {'singular_name': sn, 'id': id})

				# Don't get any sub-objects for this 
				toget.remove(id)

				# This object does not really exist on the server
				ids.remove(id)

		c(pn, "progress", message=_("Cleaning up %s which have disappeared...") % pn)

		# Remove any objects which are no longer on the server
		onserver  = set(ids)
		havelocal = set(cache().keys())
		for id in havelocal-onserver:
			c(pn, "progress", \
				message=_("Removing %(singular_name)s %(name)s as it has disappeared...") % {'singular_name': sn, 'name': cache(id).name})
			del cache()[id]

		if pn == "objects":
			c(pn, "progress", \
				message=_("Building two way tree of the universe for speed..."))
			def build(frame, parent=None, self=self):
				if parent:
					frame.parent = parent.id

				for id in frame.contains:
					try:
						build(cache(id), frame)
					except KeyError:
						from threads import NetworkThread
						raise NetworkThread.NetworkFailureEvent("%s (ID %i) references an object with ID %i, which does not exist!" % (frame.name, frame.id, id))
			build(cache(0))

		c(pn, "finished", message=_("Received all %s...") % pn)

		return toget

	def __getSubObjects(self, connection, toget, plural_name, subname, number, callback=None):
		c = callback
		pn = plural_name
		sn = plural_name[:-1]

		def cache(id=None, self=self, pn=pn):
			if id==None:
				return getattr(self, pn)
			else:
				return getattr(self, pn)[id]

		c = callback
		sb = subname

		c(sb, "start", message=_("Getting %s..") % sb)
		c(sb, "todownload", message=_("Have to get %(name)s for %(ammount)i %(plural_name)s..") % {'name': sb, 'ammount': len(toget), 'plural_name': pn}, todownload=len(toget))

		# Set the blocking so we can pipeline the requests
		connection.setblocking(True)
		empty = []
		for id in toget:
			frame = cache(id)

			if getattr(frame, number) > 0:
				c(sb, "progress", \
					message=_("Sending a request for all %(name)s on %(frame_name)s..") % {'name': sb, 'frame_name': unicode(frame.name)})
				getattr(connection, "get_%s" % sb)(id, range(0, getattr(frame, number)))
			else:
				c(sb, "progress", \
					message=_("Skipping requesting %(name)s on %(frame_name)s as there are none!") % {'name': sb, 'frame_name': unicode(frame.name)})
				empty.append(id)

		for id in empty:
			getattr(self, sb)[id] = (cache(id).modify_time, ChangeList())
			toget.remove(id)

		# Wait for the response to the order requests
		while len(toget) > 0:
			result = None
			while result is None:
				result = connection.poll()

			id = toget.pop(0)
			frame = cache(id)

			if failed(result):
				c(sb, "failure", \
					message=_("Failed to get %(name)s for %(frame_name)s (ID: %(id)s) (%(result)s)...") % {'name': sb, 'frame_name': unicode(frame.name), 'id': frame.id, 'result': result[1]})
				result = []
			else:
				c(sb, "downloaded", amount=1, \
					message=_("Got %(ammount)i %(name)s for %(frame_name)s (ID: %(id)s)...") % {'ammount': len(result), 'name': sb, 'frame_name': unicode(frame.name), 'id': frame.id})

			subs = ChangeList()
			for sub in result:
				subs.append(ChangeNode(sub))

			getattr(self, sb)[id] = (cache(id).modify_time, subs)

		c(sb, "progress", message=_("Cleaning up any stray %s..") % sb)
		for id in getattr(self, sb).keys():
			if not cache().has_key(id):
				c(sb, "progress", message=_("Found stray %(name)s for %(id)s..") % {'name': sb, 'id': id})
				del getattr(self, sb)[id]

		connection.setblocking(False)
		c(sb, "finished", message=_("Received all the %s..") % sb)

def apply(connection, evt, cache):
	"""\
	Applies a CacheDirty event to a connection.
	"""
	if evt.what == "orders":
		d = cache.orders[evt.id]

		if evt.action == "remove":
			slots = []
			for node in evt.nodes:
				assert isinstance(node, ChangeNode)
				slots.append(d.slot(node))
			slots.sort(reverse=True)

			if failed(connection.remove_orders(evt.id, slots)):
				raise IOError("Unable to remove the order...")
		
		elif evt.action in ("create after", "create before", "change"):
			assert len(evt.nodes) == 1, "%s event has multiple slots! (%r) WTF?" % (evt.action, evt.nodes)
			assert evt.change in d

			slot = d.slot(evt.change)
			if evt.action == "change":
				# Remove the old order
				if failed(connection.remove_orders(evt.id, slot)):
					raise IOError("Unable to remove the order...")

			assert not evt.change.CurrentState == "idle"
			assert not evt.change.PendingOrder is None
			if failed(connection.insert_order(evt.id, slot, evt.change.PendingOrder)):
				raise IOError("Unable to insert the order...")

			o = connection.get_orders(evt.id, slot)[0]
			if failed(o):
				raise IOError("Unable to get the order..." + o[1])

			evt.change.UpdatePending(o)

		else:
			raise SystemError("Unknown Action")

	elif evt.what == "messages" and evt.action == "remove":
		d = cache.messages[evt.id]

		slots = []
		for node in evt.nodes:
			slots.append(d.slot(node))
		slots.sort(reverse=True)

		if failed(connection.remove_messages(evt.id, slots)):
			raise IOError("Unable to remove the message...")

	elif evt.what == "designs":

		# FIXME: Assuming that these should succeed is BAD!
		if evt.action == "remove":
			if failed(connection.remove_designs(evt.change)):
				raise IOError("Unable to remove the design...")
		if evt.action == "change":
			if failed(connection.change_design(evt.change)):
				raise IOError("Unable to change the design...")
		if evt.action == "create":
			result = connection.insert_design(evt.change)
			if failed(result):
				raise IOError("Unable to add the design...")
			
			# Need to update the event with the new ID of the design.
			evt.id = result.id

	elif evt.what == "categories":

		# FIXME: Assuming that these should succeed is BAD!
		if evt.action == "remove":
			if failed(connection.remove_categories(evt.change)):
				raise IOError("Unable to remove the category...")
		if evt.action == "change":
			if failed(connection.change_category(evt.change)):
				raise IOError("Unable to change the category...")
		if evt.action == "create":
			result = connection.insert_category(evt.change)
			if failed(result):
				raise IOError("Unable to add the category...")
			
			# Need to update the event with the new ID of the design.
			evt.id = result.id
	else:
		raise ValueError("Can't deal with that yet!")

	cache.commit(evt)
	return evt



"""\
This file contains function useful for storing and retriving config.
"""

import sys
import os
import os.path

try:
	import cPickle as pickle
except ImportError:
	import pickle

def configpath():
	"""\
	Figures out where to save the preferences.
	"""
	dirs = [("APPDATA", "Thousand Parsec"), ("HOME", ".tp"), (".", "var")]
	for base, extra in dirs:
		if base in os.environ:
			base = os.environ[base]
		elif base != ".":
			continue
			
		rc = os.path.join(base, extra)
		if not os.path.exists(rc):
			os.mkdir(rc)
		return rc

def load_data(file):
	"""\
	Loads preference data from a file.
	"""
	try:
		f = open(os.path.join(configpath(), file), "r")
		data = pickle.load(f)
	except IOError:
		return None
	return data
	
def save_data(file, data):
	"""\
	Saves preference data to a file.
	"""
	f = open(os.path.join(configpath(), file), "w")
	pickle.dump(data, f)
